# Owner(s): ["oncall: quantization"]

from torch.testing._internal.common_utils import raise_on_run_directly

from .common import AOMigrationTestCase


class TestAOMigrationQuantization(AOMigrationTestCase):
    r"""Modules and functions related to the
    `torch/quantization` migration to `torch/ao/quantization`.
    """

    def test_function_import_quantize(self):
        function_list = [
            "_convert",
            "_observer_forward_hook",
            "_propagate_qconfig_helper",
            "_remove_activation_post_process",
            "_remove_qconfig",
            "_add_observer_",
            "add_quant_dequant",
            "convert",
            "_get_observer_dict",
            "_get_unique_devices_",
            "_is_activation_post_process",
            "prepare",
            "prepare_qat",
            "propagate_qconfig_",
            "quantize",
            "quantize_dynamic",
            "quantize_qat",
            "_register_activation_post_process_hook",
            "swap_module",
        ]
        self._test_function_import("quantize", function_list)

    def test_function_import_stubs(self):
        function_list = [
            "QuantStub",
            "DeQuantStub",
            "QuantWrapper",
        ]
        self._test_function_import("stubs", function_list)

    def test_function_import_quantize_jit(self):
        function_list = [
            "_check_is_script_module",
            "_check_forward_method",
            "script_qconfig",
            "script_qconfig_dict",
            "fuse_conv_bn_jit",
            "_prepare_jit",
            "prepare_jit",
            "prepare_dynamic_jit",
            "_convert_jit",
            "convert_jit",
            "convert_dynamic_jit",
            "_quantize_jit",
            "quantize_jit",
            "quantize_dynamic_jit",
        ]
        self._test_function_import("quantize_jit", function_list)

    def test_function_import_fake_quantize(self):
        function_list = [
            "_is_per_channel",
            "_is_per_tensor",
            "_is_symmetric_quant",
            "FakeQuantizeBase",
            "FakeQuantize",
            "FixedQParamsFakeQuantize",
            "FusedMovingAvgObsFakeQuantize",
            "default_fake_quant",
            "default_weight_fake_quant",
            "default_fixed_qparams_range_neg1to1_fake_quant",
            "default_fixed_qparams_range_0to1_fake_quant",
            "default_per_channel_weight_fake_quant",
            "default_histogram_fake_quant",
            "default_fused_act_fake_quant",
            "default_fused_wt_fake_quant",
            "default_fused_per_channel_wt_fake_quant",
            "_is_fake_quant_script_module",
            "disable_fake_quant",
            "enable_fake_quant",
            "disable_observer",
            "enable_observer",
        ]
        self._test_function_import("fake_quantize", function_list)

    def test_function_import_fuse_modules(self):
        function_list = [
            "_fuse_modules",
            "_get_module",
            "_set_module",
            "fuse_conv_bn",
            "fuse_conv_bn_relu",
            "fuse_known_modules",
            "fuse_modules",
            "get_fuser_method",
        ]
        self._test_function_import("fuse_modules", function_list)

    def test_function_import_quant_type(self):
        function_list = [
            "QuantType",
            "_get_quant_type_to_str",
        ]
        self._test_function_import("quant_type", function_list)

    def test_function_import_observer(self):
        function_list = [
            "_PartialWrapper",
            "_with_args",
            "_with_callable_args",
            "ABC",
            "ObserverBase",
            "_ObserverBase",
            "MinMaxObserver",
            "MovingAverageMinMaxObserver",
            "PerChannelMinMaxObserver",
            "MovingAveragePerChannelMinMaxObserver",
            "HistogramObserver",
            "PlaceholderObserver",
            "RecordingObserver",
            "NoopObserver",
            "_is_activation_post_process",
            "_is_per_channel_script_obs_instance",
            "get_observer_state_dict",
            "load_observer_state_dict",
            "default_observer",
            "default_placeholder_observer",
            "default_debug_observer",
            "default_weight_observer",
            "default_histogram_observer",
            "default_per_channel_weight_observer",
            "default_dynamic_quant_observer",
            "default_float_qparams_observer",
        ]
        self._test_function_import("observer", function_list)

    def test_function_import_qconfig(self):
        function_list = [
            "QConfig",
            "default_qconfig",
            "default_debug_qconfig",
            "default_per_channel_qconfig",
            "QConfigDynamic",
            "default_dynamic_qconfig",
            "float16_dynamic_qconfig",
            "float16_static_qconfig",
            "per_channel_dynamic_qconfig",
            "float_qparams_weight_only_qconfig",
            "default_qat_qconfig",
            "default_weight_only_qconfig",
            "default_activation_only_qconfig",
            "default_qat_qconfig_v2",
            "get_default_qconfig",
            "get_default_qat_qconfig",
            "_assert_valid_qconfig",
            "QConfigAny",
            "_add_module_to_qconfig_obs_ctr",
            "qconfig_equals",
        ]
        self._test_function_import("qconfig", function_list)

    def test_function_import_quantization_mappings(self):
        function_list = [
            "no_observer_set",
            "get_default_static_quant_module_mappings",
            "get_static_quant_module_class",
            "get_dynamic_quant_module_class",
            "get_default_qat_module_mappings",
            "get_default_dynamic_quant_module_mappings",
            "get_default_qconfig_propagation_list",
            "get_default_compare_output_module_list",
            "get_default_float_to_quantized_operator_mappings",
            "get_quantized_operator",
            "_get_special_act_post_process",
            "_has_special_act_post_process",
        ]
        dict_list = [
            "DEFAULT_REFERENCE_STATIC_QUANT_MODULE_MAPPINGS",
            "DEFAULT_STATIC_QUANT_MODULE_MAPPINGS",
            "DEFAULT_QAT_MODULE_MAPPINGS",
            "DEFAULT_DYNAMIC_QUANT_MODULE_MAPPINGS",
            # "_INCLUDE_QCONFIG_PROPAGATE_LIST",
            "DEFAULT_FLOAT_TO_QUANTIZED_OPERATOR_MAPPINGS",
            "DEFAULT_MODULE_TO_ACT_POST_PROCESS",
        ]
        self._test_function_import("quantization_mappings", function_list)
        self._test_dict_import("quantization_mappings", dict_list)

    def test_function_import_fuser_method_mappings(self):
        function_list = [
            "fuse_conv_bn",
            "fuse_conv_bn_relu",
            "fuse_linear_bn",
            "get_fuser_method",
        ]
        dict_list = ["_DEFAULT_OP_LIST_TO_FUSER_METHOD"]
        self._test_function_import("fuser_method_mappings", function_list)
        self._test_dict_import("fuser_method_mappings", dict_list)

    def test_function_import_utils(self):
        function_list = [
            "activation_dtype",
            "activation_is_int8_quantized",
            "activation_is_statically_quantized",
            "calculate_qmin_qmax",
            "check_min_max_valid",
            "get_combined_dict",
            "get_qconfig_dtypes",
            "get_qparam_dict",
            "get_quant_type",
            "get_swapped_custom_module_class",
            "getattr_from_fqn",
            "is_per_channel",
            "is_per_tensor",
            "weight_dtype",
            "weight_is_quantized",
            "weight_is_statically_quantized",
        ]
        self._test_function_import("utils", function_list)


if __name__ == "__main__":
    raise_on_run_directly("test/test_quantization.py")



# Owner(s): ["module: inductor"]
"""
OpInfo-based property tests for inductor numerical correctness.

Tests three properties:
1. Batch invariance - output shouldn't change based on batch size
2. Run-to-run determinism - same input should give same output across
   compilations, even with different autotuning choices
3. Bitwise equivalence with torch eager mode

Tests three compilation backends:
1. aot_eager_decomp_partition - AOT autograd with eager execution
2. inductor_default - Standard inductor compilation
3. inductor_numerics - Inductor with strict numerics flags

Focuses on ops commonly used in LLMs from unary_ufuncs, binary_ufuncs, and op_db.

Note: How to modify this test
-----------------------------
To add a new op:
    Add the op name to LLM_UNARY_OP_NAMES, LLM_BINARY_OP_NAMES, or LLM_OP_DB_NAMES
    depending on which OpInfo database contains the op.

To add an expected failure:
    Add to the appropriate *_XFAILS dict (e.g., EAGER_EQUIV_XFAILS):
        "backend": {"op_name": {dtype1, dtype2, ...}}
    Use ALL (None) to match all dtypes.

To remove a stale expected failure:
    Run the tests - any expected failure that now passes will fail with
    "XPASS" and tell you which entry to remove.

To run a specific test:
    pytest test/inductor/test_torchinductor_opinfo_properties.py -k "silu and eager"
"""

import sys
import unittest

import pytest

import torch
from torch._inductor.test_case import run_tests, TestCase
from torch._inductor.utils import fresh_inductor_cache
from torch.testing._internal.common_device_type import (
    instantiate_device_type_tests,
    onlyCUDA,
    ops,
)
from torch.testing._internal.common_methods_invocations import (
    binary_ufuncs,
    op_db,
    unary_ufuncs,
)
from torch.testing._internal.common_utils import (
    IS_WINDOWS,
    parametrize,
    skipIfTorchDynamo,
    TEST_WITH_ASAN,
)
from torch.testing._internal.inductor_utils import HAS_GPU


# LLM-useful op names to filter from unary_ufuncs
LLM_UNARY_OP_NAMES = {
    # Basic math
    "abs",
    "neg",
    "reciprocal",
    # Exponential/log
    "exp",
    "exp2",
    "expm1",
    "log",
    "log2",
    "log10",
    "log1p",
    # Power/root
    "sqrt",
    "rsqrt",
    # Trigonometric
    "sin",
    "cos",
    "tan",
    "tanh",
    # Activations
    "sigmoid",
    "nn.functional.relu6",
}

# LLM-useful op names to filter from binary_ufuncs
LLM_BINARY_OP_NAMES = {
    "sub",
    "mul",
    "remainder",
    "fmod",
    "maximum",
    "minimum",
}

# LLM-useful op names from op_db (nn.functional and others)
LLM_OP_DB_NAMES = {
    # Activations
    "nn.functional.gelu",
    "nn.functional.silu",
    "nn.functional.leaky_relu",
    "nn.functional.hardswish",
    # Normalization
    "nn.functional.layer_norm",
    "nn.functional.rms_norm",
    # Attention/linear
    "nn.functional.linear",
    "bmm",
    # Softmax
    "softmax",
    "log_softmax",
}

# Filter OpInfos for LLM-useful ops
llm_unary_ops = [op for op in unary_ufuncs if op.name in LLM_UNARY_OP_NAMES]
llm_binary_ops = [op for op in binary_ufuncs if op.name in LLM_BINARY_OP_NAMES]
llm_op_db_ops = [op for op in op_db if op.name in LLM_OP_DB_NAMES]

# Combine all ops, avoiding duplicates by name
_seen_names = set()
llm_ops = []
for op in llm_unary_ops + llm_binary_ops + llm_op_db_ops:
    if op.name not in _seen_names:
        _seen_names.add(op.name)
        llm_ops.append(op)

# Backends to test
BACKENDS = [
    "aot_eager_decomp_partition",
    "inductor_default",
    "inductor_numerics",
]

# Dtypes to test - common LLM dtypes
DTYPES = [
    torch.float32,
    torch.float16,
    torch.bfloat16,
]

# Number of samples for numerical testing
NUM_SAMPLES = 65536


def generate_exhaustive_16bit(dtype, device):
    """Generate all 65536 possible values for a 16-bit float dtype.

    For fp16 and bf16, there are exactly 2^16 = 65536 possible bit patterns.
    This allows exhaustive testing of unary functions.
    """
    # Generate all 16-bit patterns as uint16
    all_patterns = torch.arange(0, 65536, dtype=torch.int32, device=device)
    # View as the target dtype
    if dtype == torch.float16:
        return all_patterns.to(torch.int16).view(torch.float16)
    elif dtype == torch.bfloat16:
        return all_patterns.to(torch.int16).view(torch.bfloat16)
    else:
        raise ValueError(f"Unsupported dtype for exhaustive 16-bit testing: {dtype}")


def generate_sampled_fp32(num_samples, device, seed=42):
    """Generate uniformly sampled fp32 bit patterns.

    Samples random 32-bit patterns and interprets them as fp32.
    This provides good coverage of the fp32 space including denormals,
    infinities, and NaNs.
    """
    gen = torch.Generator(device=device)
    gen.manual_seed(seed)
    # Generate random 32-bit patterns
    patterns = torch.randint(
        0, 2**31, (num_samples,), dtype=torch.int32, device=device, generator=gen
    )
    # Also include negative patterns (high bit set)
    signs = torch.randint(
        0, 2, (num_samples,), dtype=torch.int32, device=device, generator=gen
    )
    patterns = patterns | (signs << 31)
    return patterns.view(torch.float32)


def generate_sampled_pairs(num_samples, dtype, device, seed=42):
    """Generate pairs of sampled values for binary function testing.

    For fp32: uses random bit patterns
    For fp16/bf16: samples from all possible values
    """
    gen = torch.Generator(device=device)
    gen.manual_seed(seed)

    if dtype == torch.float32:
        # Random bit patterns for fp32
        patterns1 = torch.randint(
            0, 2**31, (num_samples,), dtype=torch.int32, device=device, generator=gen
        )
        signs1 = torch.randint(
            0, 2, (num_samples,), dtype=torch.int32, device=device, generator=gen
        )
        patterns1 = patterns1 | (signs1 << 31)
        x = patterns1.view(torch.float32)

        patterns2 = torch.randint(
            0, 2**31, (num_samples,), dtype=torch.int32, device=device, generator=gen
        )
        signs2 = torch.randint(
            0, 2, (num_samples,), dtype=torch.int32, device=device, generator=gen
        )
        patterns2 = patterns2 | (signs2 << 31)
        y = patterns2.view(torch.float32)
    else:
        # Sample indices into the 16-bit space
        indices1 = torch.randint(
            0, 65536, (num_samples,), dtype=torch.int32, device=device, generator=gen
        )
        indices2 = torch.randint(
            0, 65536, (num_samples,), dtype=torch.int32, device=device, generator=gen
        )

        if dtype == torch.float16:
            x = indices1.to(torch.int16).view(torch.float16)
            y = indices2.to(torch.int16).view(torch.float16)
        elif dtype == torch.bfloat16:
            x = indices1.to(torch.int16).view(torch.bfloat16)
            y = indices2.to(torch.int16).view(torch.bfloat16)
        else:
            raise ValueError(f"Unsupported dtype: {dtype}")

    return x, y


# Inductor options for strict numerics matching eager behavior
INDUCTOR_NUMERICS_OPTIONS = {
    "deterministic": True,
    "fallback_random": True,
    "emulate_precision_casts": True,
    "eager_numerics.division_rounding": True,
    "eager_numerics.disable_ftz": True,
    "eager_numerics.use_pytorch_libdevice": True,
}


def _slice_tensor_or_collection(value, batch_size, original_batch_size, input_ndim):
    """Recursively slice tensors in a value (tensor, list, tuple, or other).

    Only slices tensors that have the same number of dimensions as the input
    to preserve broadcasting patterns.
    """
    if isinstance(value, torch.Tensor):
        if (
            value.dim() == input_ndim  # Same ndim to preserve broadcast pattern
            and value.dim() > 0
            and value.shape[0] == original_batch_size
            and value.shape[0] > 1  # Don't slice broadcast dimensions
        ):
            return value[:batch_size]
        return value
    elif isinstance(value, (list, tuple)):
        sliced = [
            _slice_tensor_or_collection(v, batch_size, original_batch_size, input_ndim)
            for v in value
        ]
        return type(value)(sliced)
    else:
        return value


def slice_tensors_to_batch_size(sample_input, batch_size):
    """Slice all tensors in a SampleInput to the given batch size.

    For batch invariance testing, we need to slice all tensor inputs that have
    matching batch dimensions. This function slices:
    - sample_input.input if it's a tensor with dim > 0
    - All tensor args that have the same size in dim 0 as the input (not broadcast dims)
    - All tensor kwargs that have the same size in dim 0 as the input (not broadcast dims)

    Tensors with size 1 in dim 0 are broadcast dimensions and should not be sliced.
    Also handles nested lists/tuples of tensors (e.g., TensorList).
    Only slices tensors with the same ndim as the input to preserve broadcast patterns.

    Returns a tuple (sliced_input, sliced_args, sliced_kwargs), or None if slicing is not possible.
    """
    inp = sample_input.input
    if not isinstance(inp, torch.Tensor) or inp.dim() == 0:
        return None

    original_batch_size = inp.shape[0]
    input_ndim = inp.dim()
    if batch_size > original_batch_size:
        return None

    # Slice the input
    sliced_input = inp[:batch_size]

    # Slice args - handles tensors and lists/tuples of tensors
    sliced_args = tuple(
        _slice_tensor_or_collection(arg, batch_size, original_batch_size, input_ndim)
        for arg in sample_input.args
    )

    # Slice kwargs - handles tensors and lists/tuples of tensors
    sliced_kwargs = {
        key: _slice_tensor_or_collection(
            val, batch_size, original_batch_size, input_ndim
        )
        for key, val in sample_input.kwargs.items()
    }

    return sliced_input, sliced_args, sliced_kwargs


def sample_operates_on_batch_dim(op_name, sample_input):
    """Check if a sample input operates on the batch dimension (dim 0).

    For ops that normalize/reduce over a dimension, if that dimension is 0,
    slicing the batch will change the result and batch invariance doesn't apply.
    """
    # Ops that take a 'dim' argument and normalize/reduce over it
    dim_based_ops = {
        "softmax",
        "log_softmax",
        "nn.functional.softmax",
        "nn.functional.log_softmax",
    }

    if op_name not in dim_based_ops:
        return False

    # Get dim from args or kwargs
    dim = None
    if sample_input.args:
        dim = sample_input.args[0]
    if "dim" in sample_input.kwargs:
        dim = sample_input.kwargs["dim"]

    # If dim is 0 or -ndim (equivalent to 0), the op operates on the batch dimension
    if dim is not None:
        inp = sample_input.input
        if isinstance(inp, torch.Tensor):
            ndim = inp.dim()
            # Normalize negative dim
            if dim < 0:
                dim = dim + ndim
            return dim == 0

    return False


# Expected failures for bitwise equivalence tests.
# Structure: backend -> op_name -> set of failing dtypes (None means all dtypes)
# These track known numerical differences between eager and compiled execution.
# The goal is to eventually fix these and remove entries.

# Dtype shorthands
fp32, fp16, bf16, ALL = torch.float32, torch.float16, torch.bfloat16, None

EAGER_EQUIV_XFAILS = {
    "aot_eager_decomp_partition": {
        "nn.functional.gelu": {fp32},
        "nn.functional.layer_norm": {fp32},
        "nn.functional.rms_norm": {fp32},
        "softmax": {fp32},
        "log_softmax": {fp32},
    },
    "inductor_default": {
        "reciprocal": {fp32},
        "remainder": {ALL},
        "sigmoid": {fp32},
        "nn.functional.gelu": {fp32},
        "nn.functional.layer_norm": {fp32},
        "nn.functional.silu": {fp16, fp32},
        "softmax": {fp32},
        "log_softmax": {fp32},
    },
    "inductor_numerics": {
        "remainder": {ALL},
        "sigmoid": {fp32},
        "sub": {ALL},
        "nn.functional.gelu": {fp32},
        "nn.functional.layer_norm": {fp32},
        "softmax": {fp32},
        "log_softmax": {fp32},
    },
}

DETERMINISM_XFAILS = {}

BATCH_INVARIANCE_XFAILS = {
    "aot_eager_decomp_partition": {
        "nn.functional.linear": {ALL},
    },
    "inductor_default": {
        "nn.functional.linear": {ALL},
    },
    "inductor_numerics": {
        "nn.functional.linear": {ALL},
    },
}

UNARY_NUMERICAL_XFAILS = {
    "inductor_default": {
        "exp2": {bf16, fp32},
        "expm1": {bf16, fp32},
        "log1p": {fp32},
        "reciprocal": {fp32},
        "rsqrt": {bf16, fp32},
        "sigmoid": {fp32},
        "sin": {fp32},
        "tan": {bf16, fp32},
        "tanh": {fp32},
    },
    "inductor_numerics": {
        "sigmoid": {fp32},
    },
}

BINARY_NUMERICAL_XFAILS = {
    "inductor_default": {
        "fmod": {bf16, fp32},
        "remainder": {ALL},
    },
    "inductor_numerics": {
        "remainder": {ALL},
    },
}

XFAIL_DICTS = {
    "eager_equivalence": EAGER_EQUIV_XFAILS,
    "determinism": DETERMINISM_XFAILS,
    "batch_invariance": BATCH_INVARIANCE_XFAILS,
    "unary_numerical": UNARY_NUMERICAL_XFAILS,
    "binary_numerical": BINARY_NUMERICAL_XFAILS,
}

ROCM_EAGER_EQUIV_XFAILS = {
    "aot_eager_decomp_partition": {
        "nn.functional.gelu": {fp32},
        "nn.functional.layer_norm": {fp32},
        "nn.functional.rms_norm": {fp32},
        "softmax": {fp32},
        "log_softmax": {fp32},
    },
    "inductor_default": {
        "sigmoid": {fp32},
        "nn.functional.gelu": {fp32},
        "nn.functional.layer_norm": {fp32},
        "nn.functional.silu": {fp16, fp32},
        "softmax": {fp32},
        "log_softmax": {fp32},
    },
    "inductor_numerics": {
        "sigmoid": {fp32},
        "sub": {ALL},
        "nn.functional.gelu": {fp32},
        "nn.functional.layer_norm": {fp32},
        "softmax": {fp32},
        "log_softmax": {fp32},
    },
}

ROCM_DETERMINISM_XFAILS = {}

ROCM_BATCH_INVARIANCE_XFAILS = {
    "aot_eager_decomp_partition": {
        "nn.functional.linear": {ALL},
    },
    "inductor_default": {
        "nn.functional.linear": {ALL},
        "log1p": {fp32},
    },
    "inductor_numerics": {
        "nn.functional.linear": {ALL},
    },
}

ROCM_UNARY_NUMERICAL_XFAILS = {
    "inductor_default": {
        "log1p": {fp32},
        "rsqrt": {bf16, fp32},
        "sigmoid": {fp32},
        "sin": {fp32},
        "tan": {fp32},
        "tanh": {fp32},
    },
    "inductor_numerics": {
        "sigmoid": {fp32},
    },
}

ROCM_BINARY_NUMERICAL_XFAILS = {}

ROCM_XFAIL_DICTS = {
    "eager_equivalence": ROCM_EAGER_EQUIV_XFAILS,
    "determinism": ROCM_DETERMINISM_XFAILS,
    "batch_invariance": ROCM_BATCH_INVARIANCE_XFAILS,
    "unary_numerical": ROCM_UNARY_NUMERICAL_XFAILS,
    "binary_numerical": ROCM_BINARY_NUMERICAL_XFAILS,
}


def is_expected_failure(device_type, op_name, backend, test_type, dtype=None):
    """Check if a test is expected to fail."""
    xfail_dicts = ROCM_XFAIL_DICTS if torch.version.hip is not None else XFAIL_DICTS
    xfails = xfail_dicts.get(test_type, {}).get(backend, {}).get(op_name, set())
    return dtype in xfails or ALL in xfails


def compile_fn(fn, backend):
    """Compile a function with the given backend."""
    if backend == "aot_eager_decomp_partition":
        return torch.compile(fn, backend="aot_eager_decomp_partition")
    elif backend == "inductor_default":
        return torch.compile(fn, backend="inductor")
    elif backend == "inductor_numerics":
        return torch.compile(fn, backend="inductor", options=INDUCTOR_NUMERICS_OPTIONS)
    else:
        raise ValueError(f"Unknown backend: {backend}")


@unittest.skipIf(IS_WINDOWS, "Skipped on Windows")
@unittest.skipIf(TEST_WITH_ASAN, "Skipped under ASAN")
@unittest.skipIf(not HAS_GPU, "Requires GPU")
class TestOpInfoProperties(TestCase):
    """Test op properties under various inductor modes using OpInfo on CUDA."""

    def setUp(self):
        super().setUp()
        torch._dynamo.reset()
        patch = torch._dynamo.config.patch(
            recompile_limit=sys.maxsize, accumulated_recompile_limit=sys.maxsize
        )
        patch.__enter__()
        self.addCleanup(patch.__exit__, None, None, None)

    def tearDown(self):
        super().tearDown()
        torch._dynamo.reset()

    def _get_sample_inputs(self, op, device, dtype):
        """Get sample inputs from OpInfo using reference_inputs for comprehensive coverage."""
        # Use reference_inputs for more comprehensive test coverage
        # Falls back to sample_inputs if reference_inputs_func is not defined
        try:
            samples = list(op.reference_inputs(device, dtype, requires_grad=False))
        except Exception:
            samples = list(op.sample_inputs(device, dtype, requires_grad=False))
        return samples

    def _run_with_expected_failure(
        self, device_type, op_name, backend, test_type, dtype, test_fn
    ):
        """Run a test with expected failure handling.

        Uses pytest.xfail for clear test output:
        - If test is expected to fail and does fail: XFAIL (pytest.xfail called)
        - If test is expected to fail but passes: FAILED (strict xpass behavior)
        - If test is not expected to fail: runs normally (PASSED or FAILED)

        Args:
            test_fn: A callable that runs the actual test assertions
        """
        expected = is_expected_failure(device_type, op_name, backend, test_type, dtype)

        if expected:
            try:
                test_fn()
            except (AssertionError, RuntimeError):
                pytest.xfail(f"Known failure: {op_name}/{backend}/{dtype}")
            else:
                self.fail(
                    f"XPASS: {op_name}/{backend}/{dtype} - remove from {test_type} xfails"
                )
        else:
            test_fn()

    # =========================================================================
    # Batch Invariance Tests
    # =========================================================================

    @onlyCUDA
    @skipIfTorchDynamo("Test uses dynamo already")
    @ops(llm_ops, allowed_dtypes=DTYPES)
    @parametrize("backend", BACKENDS)
    def test_batch_invariance(self, device, dtype, op, backend):
        """Test batch invariance with exponentially decreasing batch sizes.

        For each sample input, this test:
        1. Runs the compiled op at full batch size
        2. Runs at size/2, size/4, etc. (exponentially decreasing)
        3. Verifies the sliced output matches the corresponding slice of the full output

        All tensor inputs with matching batch dimensions are sliced together.

        Note: We disable split-k GEMM accumulation to ensure consistent results
        across batch sizes, as split-k can produce different rounding based on
        how the work is partitioned.
        """
        torch._dynamo.reset()
        device_type = torch.device(device).type

        samples = self._get_sample_inputs(op, device, dtype)
        if not samples:
            self.skipTest(f"No samples for {op.name}")

        fn = op.get_op()

        # Disable split-k accumulation in GEMM to ensure batch-invariant results.
        # Split-k can produce different rounding based on how work is partitioned.
        # Disabling split-k requires the cuBLASLt backend, so we set it first.
        prev_blas_library = torch.backends.cuda.preferred_blas_library()
        torch.backends.cuda.preferred_blas_library("cublaslt")

        # We need to save both the reduced precision and split_k settings.
        prev_fp16 = (
            torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction,
            torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction_split_k,
        )
        prev_bf16 = (
            torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction,
            torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction_split_k,
        )
        # Disable both reduced precision and split-k
        torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction = (
            False,
            False,
        )
        torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction = (
            False,
            False,
        )

        def run_test():
            tested_any = False

            for sample in samples:
                # Skip if input is not a tensor or is 0-dim
                if (
                    not isinstance(sample.input, torch.Tensor)
                    or sample.input.dim() == 0
                ):
                    continue

                # Need at least size 4 in first dimension for meaningful test
                full_size = sample.input.shape[0]
                if full_size < 4:
                    continue

                # Skip broadcast/expanded tensors (stride 0 in batch dim)
                # These don't have meaningful batch invariance since all rows are the same
                if sample.input.stride()[0] == 0:
                    continue

                # Skip samples where the op normalizes/reduces over dim 0 (batch dimension)
                # because slicing the batch changes the normalization result
                if sample_operates_on_batch_dim(op.name, sample):
                    continue

                compiled_fn = compile_fn(fn, backend)

                # Get reference output at full size
                full_args = (sample.input,) + tuple(sample.args)
                full_kwargs = sample.kwargs
                full_out = compiled_fn(*full_args, **full_kwargs)

                if not isinstance(full_out, torch.Tensor):
                    continue

                # Skip if output is 0-dim (scalar) - can't slice it
                if full_out.dim() == 0:
                    continue

                # Skip if output's first dimension doesn't match input's batch size
                # (e.g., due to broadcasting or reduction)
                if full_out.shape[0] != full_size:
                    continue

                tested_any = True

                # Test with exponentially decreasing sizes: size, size/2, size/4, ...
                size = full_size
                while size >= 1:
                    # Slice all tensor inputs with matching batch dimensions
                    sliced = slice_tensors_to_batch_size(sample, size)
                    if sliced is None:
                        break
                    sliced_input, sliced_args, sliced_kwargs = sliced

                    out = compiled_fn(sliced_input, *sliced_args, **sliced_kwargs)

                    # Verify output matches the corresponding slice of full output (bitwise)
                    self.assertEqual(out, full_out[:size], rtol=0, atol=0)

                    # Step down exponentially
                    size = size // 2

            if not tested_any:
                self.skipTest("No suitable samples found for batch invariance test")

        try:
            self._run_with_expected_failure(
                device_type, op.name, backend, "batch_invariance", dtype, run_test
            )
        finally:
            torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction = (
                prev_fp16
            )
            torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction = (
                prev_bf16
            )
            torch.backends.cuda.preferred_blas_library(prev_blas_library)

    # =========================================================================
    # Run-to-Run Determinism Tests
    # =========================================================================

    @onlyCUDA
    @skipIfTorchDynamo("Test uses dynamo already")
    @ops(llm_ops, allowed_dtypes=DTYPES)
    @parametrize("backend", BACKENDS)
    def test_determinism(self, device, dtype, op, backend):
        """Test run-to-run determinism with fresh cache and benchmark perturbation.

        For enhanced determinism testing, this test:
        1. Resets dynamo and uses fresh inductor cache between compilations
        2. Uses distort_benchmarking_result config to perturb autotuning choices
        3. Enables deterministic algorithms to ensure reliable determinism
        4. Tests multiple runs and verifies bitwise identical outputs
        """
        torch._dynamo.reset()
        device_type = torch.device(device).type

        samples = self._get_sample_inputs(op, device, dtype)
        if not samples:
            self.skipTest(f"No samples for {op.name}")

        fn = op.get_op()

        # Enable deterministic algorithms for reliable determinism testing
        prev_deterministic = torch.are_deterministic_algorithms_enabled()
        torch.use_deterministic_algorithms(True)

        def run_test():
            import torch._inductor.config as inductor_config

            for sample in samples:
                args = (sample.input,) + tuple(sample.args)
                kwargs = sample.kwargs

                # Run 1: Fresh cache, normal benchmarking
                torch._dynamo.reset()
                with fresh_inductor_cache():
                    compiled_fn1 = compile_fn(fn, backend)
                    out1 = compiled_fn1(*args, **kwargs)

                # Run 2: Fresh cache, inverted benchmark results
                torch._dynamo.reset()
                with fresh_inductor_cache():
                    with inductor_config.patch(
                        {"test_configs.distort_benchmarking_result": "inverse"}
                    ):
                        compiled_fn2 = compile_fn(fn, backend)
                        out2 = compiled_fn2(*args, **kwargs)

                # Run 3: Fresh cache, random benchmark results
                torch._dynamo.reset()
                with fresh_inductor_cache():
                    with inductor_config.patch(
                        {"test_configs.distort_benchmarking_result": "random"}
                    ):
                        compiled_fn3 = compile_fn(fn, backend)
                        out3 = compiled_fn3(*args, **kwargs)

                # Bitwise identical
                self.assertEqual(out1, out2, rtol=0, atol=0)
                self.assertEqual(out2, out3, rtol=0, atol=0)

        try:
            self._run_with_expected_failure(
                device_type, op.name, backend, "determinism", dtype, run_test
            )
        finally:
            torch.use_deterministic_algorithms(prev_deterministic)

    # =========================================================================
    # Bitwise Equivalence with Eager Mode Tests
    # =========================================================================

    @onlyCUDA
    @skipIfTorchDynamo("Test uses dynamo already")
    @ops(llm_ops, allowed_dtypes=DTYPES)
    @parametrize("backend", BACKENDS)
    def test_eager_equivalence(self, device, dtype, op, backend):
        """Test bitwise equivalence with eager execution."""
        torch._dynamo.reset()
        device_type = torch.device(device).type

        samples = self._get_sample_inputs(op, device, dtype)
        if not samples:
            self.skipTest(f"No samples for {op.name}")

        fn = op.get_op()

        def run_test():
            for sample in samples:
                args = (sample.input,) + tuple(sample.args)
                kwargs = sample.kwargs

                # Eager reference
                eager_out = fn(*args, **kwargs)

                # Compiled output
                compiled_fn = compile_fn(fn, backend)
                compiled_out = compiled_fn(*args, **kwargs)

                # Bitwise identical
                self.assertEqual(eager_out, compiled_out, rtol=0, atol=0)

        self._run_with_expected_failure(
            device_type, op.name, backend, "eager_equivalence", dtype, run_test
        )

    # =========================================================================
    # Exhaustive/Sampled Unary Ufunc Tests
    # =========================================================================

    @onlyCUDA
    @skipIfTorchDynamo("Test uses dynamo already")
    @ops(llm_unary_ops, allowed_dtypes=DTYPES)
    @parametrize("backend", BACKENDS)
    def test_unary_ufunc_numerical(self, device, dtype, op, backend):
        """Test unary ufuncs with exhaustive (16-bit) or sampled (fp32) inputs.

        For fp16 and bf16: exhaustively tests all 65536 possible bit patterns.
        For fp32: tests 64k sampled random bit patterns.

        Verifies bitwise equivalence between eager and compiled execution.
        """
        torch._dynamo.reset()
        device_type = torch.device(device).type

        fn = op.get_op()

        # Generate test values based on dtype
        if dtype in (torch.float16, torch.bfloat16):
            # Exhaustive: all 65536 possible values
            test_values = generate_exhaustive_16bit(dtype, device)
        else:
            # Sampled: 64k random bit patterns
            test_values = generate_sampled_fp32(NUM_SAMPLES, device)

        # Eager reference
        eager_out = fn(test_values)

        # Compiled output
        compiled_fn = compile_fn(fn, backend)
        compiled_out = compiled_fn(test_values)

        def run_test():
            # Bitwise identical
            self.assertEqual(eager_out, compiled_out, rtol=0, atol=0)

        self._run_with_expected_failure(
            device_type, op.name, backend, "unary_numerical", dtype, run_test
        )

    # =========================================================================
    # Sampled Binary Ufunc Tests
    # =========================================================================

    @onlyCUDA
    @skipIfTorchDynamo("Test uses dynamo already")
    @ops(llm_binary_ops, allowed_dtypes=DTYPES)
    @parametrize("backend", BACKENDS)
    def test_binary_ufunc_numerical(self, device, dtype, op, backend):
        """Test binary ufuncs on 64k sampled value pairs.

        For fp32: samples random 32-bit patterns.
        For fp16/bf16: samples from all possible 16-bit values.

        Verifies bitwise equivalence between eager and compiled execution.
        """
        torch._dynamo.reset()
        device_type = torch.device(device).type

        fn = op.get_op()

        # Generate sampled pairs
        x, y = generate_sampled_pairs(NUM_SAMPLES, dtype, device)

        # Eager reference
        eager_out = fn(x, y)

        # Compiled output
        compiled_fn = compile_fn(fn, backend)
        compiled_out = compiled_fn(x, y)

        def run_test():
            # Bitwise identical
            self.assertEqual(eager_out, compiled_out, rtol=0, atol=0)

        self._run_with_expected_failure(
            device_type, op.name, backend, "binary_numerical", dtype, run_test
        )


instantiate_device_type_tests(TestOpInfoProperties, globals(), except_for=["cpu"])

if __name__ == "__main__":
    run_tests()



# Copyright (c) Meta Platforms, Inc. and affiliates
# Owner(s): ["oncall: distributed"]
import itertools

import torch
from torch.distributed.tensor import distribute_tensor, DTensor, Replicate, Shard
from torch.distributed.tensor._dtensor_spec import DTensorSpec
from torch.distributed.tensor.experimental import register_sharding
from torch.testing._internal.common_utils import run_tests
from torch.testing._internal.distributed._tensor.common_dtensor import (
    DTensorTestBase,
    with_comms,
)


aten = torch.ops.aten


class TestRegisterSharding(DTensorTestBase):
    def tearDown(self):
        super().tearDown()
        # Clean up any custom ops registered during tests to avoid test pollution
        # This addresses the concern about proper deregistration of custom ops
        import torch._custom_op

        keys = list(torch._custom_op.impl.global_registry.keys())
        for key in keys:
            if key.startswith("test_dtensor::"):
                torch._custom_op.impl.global_registry[key]._destroy()
        if hasattr(torch.ops, "test_dtensor"):
            delattr(torch.ops, "test_dtensor")

    @with_comms
    def test_softmax_fwd(self):
        # After registering the custom softmax sharding strategy,
        # the original entry would have been replaced.
        # The following line is for showcasing purpose only.
        DTensor._op_dispatcher.sharding_propagator.op_strategy_funcs.pop(
            aten._softmax.default, None
        )

        @register_sharding(aten._softmax.default)
        def custom_softmax_sharding(
            x: DTensorSpec,
            dim: int,
            half_to_float: torch.dtype,
        ):
            softmax_dim = dim if dim >= 0 else dim + x.ndim

            acceptable_shardings = []

            all_replicate = ([Replicate()], [Replicate(), None, None])
            acceptable_shardings.append(all_replicate)

            for sharding_dim in range(x.ndim):
                if sharding_dim != softmax_dim:
                    all_sharded = (
                        [Shard(sharding_dim)],
                        [Shard(sharding_dim), None, None],
                    )
                    acceptable_shardings.append(all_sharded)

            return acceptable_shardings

        # check if the RuntimeSchemaInfo is derived correctly
        schema_info = DTensor._op_dispatcher.sharding_propagator.op_to_schema_info[
            aten._softmax.default
        ]
        self.assertEqual(schema_info.static_argnum, 1)

        device_mesh = self.build_device_mesh()

        x = torch.rand(8, 12, 16, device=self.device_type)
        dims = range(3)  # used to convert -1 to the actual dim
        softmax_dims = [-1, 0, 1]
        shard_dims = [0, 1, 2]
        test_list = list(itertools.product(softmax_dims, shard_dims))

        for softmax_dim, shard_dim in test_list:
            local_y = torch.nn.functional.softmax(
                x, dim=softmax_dim, dtype=torch.float32
            )
            dist_x = distribute_tensor(x, device_mesh, [Shard(shard_dim)])
            dist_y = torch.nn.functional.softmax(
                dist_x, dim=softmax_dim, dtype=torch.float32
            )
            if dims[shard_dim] == dims[softmax_dim]:
                self.assertTrue(dist_y.placements[0].is_replicate())
                self.assertEqual(dist_y.to_local(), local_y)
            else:
                self.assertTrue(dist_y.placements[0].is_shard(dim=shard_dim))
                self.assertEqual(dist_y.full_tensor(), local_y)

    @with_comms
    def test_argmax(self):
        @register_sharding(aten.argmax.default)
        def custom_argmax_sharding(x, dim, keepdim):
            acceptable_shardings = []

            all_replicate = ([Replicate()], [Replicate(), None, None])
            acceptable_shardings.append(all_replicate)

            if keepdim:
                for sharding_dim in range(x.ndim):
                    if sharding_dim != dim:
                        all_sharded = (
                            [Shard(sharding_dim)],
                            [Shard(sharding_dim), None, None],
                        )
                        acceptable_shardings.append(all_sharded)

            return acceptable_shardings

        # check if the RuntimeSchemaInfo is derived correctly
        # when the first int arg is optional
        schema_info = DTensor._op_dispatcher.sharding_propagator.op_to_schema_info[
            aten.argmax.default
        ]
        self.assertEqual(schema_info.static_argnum, 1)

        device_mesh = self.build_device_mesh()

        x = torch.rand(8, 12, device=self.device_type)
        dist_x = distribute_tensor(x, device_mesh, [Shard(0)])

        local_y = torch.argmax(x, dim=1, keepdim=True)
        dist_y = torch.argmax(dist_x, dim=1, keepdim=True)

        self.assertTrue(dist_y.placements[0].is_shard(dim=0))
        self.assertEqual(dist_y.full_tensor(), local_y)

    @with_comms
    def test_register_sharding_for_tensor_kwargs(self):
        mesh = self.build_device_mesh()
        x = torch.randn(4, 4, device=self.device_type)
        x_dt = distribute_tensor(x, mesh, [Replicate()])

        @register_sharding(aten.min.dim_min)
        def min_dim_strategy(x, dim, keepdim, min, min_indices):
            all_replicate = (
                [Replicate(), Replicate()],
                [Replicate(), None, None, Replicate(), Replicate()],
            )
            return [all_replicate]

        value = torch.randn(4, 1, device=self.device_type)
        indices = torch.randn(4, 1, device=self.device_type).long()
        value_dt = distribute_tensor(value, mesh, [Replicate()])
        indices_dt = distribute_tensor(indices, mesh, [Replicate()])

        result = torch.min(x_dt, dim=1, keepdim=True, out=(value_dt, indices_dt))

        self.assertIsInstance(result[0], DTensor)
        self.assertIsInstance(result[1], DTensor)

        expected_values, expected_indices = torch.min(x, dim=1, keepdim=True)
        self.assertEqual(result[0].full_tensor(), expected_values)
        self.assertEqual(result[1].full_tensor(), expected_indices)

    @with_comms
    def test_register_sharding_non_tensor_first_arg(self):
        # Test that get_mesh_from_args works when the first arg is not a DTensor
        # Regression test for https://github.com/pytorch/pytorch/issues/167915
        @torch.library.custom_op("test_dtensor::scalar_first_arg", mutates_args=())
        def scalar_first_arg(name: str, x: torch.Tensor) -> torch.Tensor:
            return x.clone()

        @scalar_first_arg.register_fake
        def _(name, x):
            return x.clone()

        @register_sharding(torch.ops.test_dtensor.scalar_first_arg.default)
        def scalar_first_arg_sharding(name, x):
            return [([Replicate()], [None, Replicate()])]

        mesh = self.build_device_mesh()
        x = torch.randn(4, 4, device=self.device_type)
        x_dt = distribute_tensor(x, mesh, [Replicate()])

        # This should not raise "Cannot find device mesh from args"
        result = torch.ops.test_dtensor.scalar_first_arg("test", x_dt)

        self.assertIsInstance(result, DTensor)
        self.assertEqual(result.full_tensor(), x)


if __name__ == "__main__":
    run_tests()



# Owner(s): ["oncall: jit"]
# ruff: noqa: F841

import io
import os
import sys
import unittest
from enum import Enum
from textwrap import dedent
from typing import Dict, List, Optional, Tuple, Union

import torch
from torch.testing import FileCheck


# Make the helper files in test/ importable
pytorch_test_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.append(pytorch_test_dir)
from torch.testing._internal.common_utils import raise_on_run_directly
from torch.testing._internal.jit_utils import JitTestCase, make_global


class TestUnion(JitTestCase):
    """
    This class tests the functionality of `Union`.

    Note: It's important to be able to refine the type of a `Union` to
    one of its internal types. Currently, there are differences in the
    way Python expects `isinstance` checks and the way TorchScript
    expects `isinstance` checks. This means that we can't use
    `checkScript` in our test cases because either the eager mode or the
    script mode wouldn't run! So, some test cases have separate but
    equivalent functions to emulate `checkScript`.
    """

    def test_check_union_annotation(self):
        def test_func(a: int | float, b: Optional[int]):
            return 0

        scripted_func = torch.jit.script(test_func)
        graph_rep = str(scripted_func.graph)
        code_rep = str(scripted_func.code)
        # TS graph IR for Union should be annotated as Union()
        FileCheck().check("Union(").check("int?").run(graph_rep)
        # Serialized code for Union should be annotated as Union[]
        FileCheck().check("Union[").check("Optional[int]").run(code_rep)
        self.checkScript(test_func, (5, 6))
        # this shouldn't error out
        torch._C.parse_ir(str(scripted_func.graph))

    def test_union_with_scalar_values(self):
        def fn(x: int | float) -> str:
            return "foo"

        self.checkScript(fn, (1,))
        self.checkScript(fn, (1.0,))

        scripted = torch.jit.script(fn)

        with self.assertRaisesRegex(
            RuntimeError,
            "Expected a member of"
            r" Union\[float, int\] but "
            "instead found type str",
        ):
            scripted("1")

    def test_union_with_collections(self):
        def fn(x: Dict[str, int] | List[int]) -> str:
            return "foo"

        self.checkScript(fn, ({"foo": 1, "bar": 2, "baz": 3},))
        self.checkScript(fn, ([1, 2, 3],))

        scripted = torch.jit.script(fn)

        with self.assertRaisesRegex(
            RuntimeError,
            "Expected a member of"
            r" Union\[List\[int\], Dict\[str, "
            r"int\]\] but instead found type "
            r"Dict\[str, str\]",
        ):
            scripted({"foo": "bar", "baz": "qux"})

        with self.assertRaisesRegex(
            RuntimeError,
            "Expected a member of"
            r" Union\[List\[int\], Dict\[str, "
            r"int\]\] but instead found type "
            r"List\[str\]",
        ):
            scripted(["foo", "bar", "baz"])

        with self.assertRaisesRegex(
            RuntimeError,
            "Expected a member of"
            r" Union\[List\[int\], Dict\[str, "
            r"int\]\] but instead found type "
            "str",
        ):
            scripted("1")

    def test_union_with_enum(self):
        class Color(Enum):
            RED = 1
            GREEN = 2

        make_global(Color)

        def fn(x: str | Color) -> str:
            return "foo"

        self.checkScript(fn, (Color.RED,))
        self.checkScript(fn, ("red",))

        scripted = torch.jit.script(fn)

        with self.assertRaisesRegex(
            RuntimeError,
            "Expected a member of"
            r" Union\[__torch__.jit.test_union_pep604."
            r"Color, str\] but instead found "
            "type int",
        ):
            scripted(1)

    def test_union_in_class_constructor(self):
        @torch.jit.script  # noqa: B903
        class A:  # noqa: B903
            def __init__(self, x: int | str) -> None:
                self.x = x

        def fn(x: str | int) -> A:
            return A(x)

        self.assertEqual(fn("foo").x, "foo")
        self.assertEqual(fn(1).x, 1)

        scripted = torch.jit.script(fn)

        with self.assertRaisesRegex(
            RuntimeError,
            "Expected a member of"
            r" Union\[int, str\] but instead "
            r"found type List\[str\]",
        ):
            scripted(["foo", "bar", "baz"])

    def test_union_return_type(self):
        def fn(x: int) -> int | str:
            return "foo"

        self.checkScript(fn, (1,))

    def test_union_as_annotation(self):
        def fn() -> int | str:
            x: int | str = "foo"
            return x

        self.checkScript(fn, ())

    def test_union_as_annotation_in_typed_container(self):
        def fn() -> None:
            l: List[int | str] = []
            u1: int | str = "foo"
            u2: int | str = 1
            l.append(u1)
            l.append(u2)

        self.checkScript(fn, ())

    def test_union_as_annotation_py2(self):
        def fn():
            # type: () -> int | str
            x: int | str = "foo"
            return x

        self.checkScript(fn, ())

    def test_union_as_internal_tuple_type(self):
        def fn():
            t: Tuple[int | str, int | str] = (1, "foo")
            return t

        self.checkScript(fn, ())

    def test_union_variable_can_be_reassigned(self):
        @torch.jit.script
        def aux1(i: int):
            return int(i**2)

        @torch.jit.script
        def aux2(s: str):
            return s + s

        def fn() -> int | str:
            x: int | str = "foo"
            i: int = 1
            x = i
            y: int = aux1(x)
            z: str = aux2(str(y))
            x = z
            return x

        self.checkScript(fn, ())

    def test_union_does_not_replace_existing_annotated_type(self):
        def fn():
            x: List[int] = [1, 2, 3]
            x.append("foo")
            return x

        with self.assertRaisesRegex(RuntimeError, "Could not match type str"):
            scripted = torch.jit.script(fn)
            scripted()

    def test_union_does_not_replace_existing_annotated_type_union(self):
        def fn():
            x: List[int | str] = [1, "foo", 3]
            x.append(2.0)
            return x

        with self.assertRaisesRegex(RuntimeError, "Could not match type float"):
            scripted = torch.jit.script(fn)
            scripted()

    def test_union_does_not_replace_existing_annotated_type_empty_container(self):
        def fn():
            x: List[int] = []
            x.append("foo")
            return x

        with self.assertRaisesRegex(RuntimeError, "Could not match type str"):
            scripted = torch.jit.script(fn)
            scripted()

    def test_unions_of_unions_are_flattened(self):
        @torch.jit.script
        def fn(x: (int | str) | float) -> str:
            return "foo"

        s = fn.graph

        FileCheck().check("x : Union(float, int, str)").run(s)

    def test_unions_of_a_single_argument_vanish(self):
        @torch.jit.script
        def fn(x: Union[int]) -> str:
            return "foo"

        s = fn.graph

        FileCheck().check("x : int").run(s)

    def test_union_redundant_arguments_are_skipped(self):
        @torch.jit.script
        def fn(x: int | str | int) -> str:
            return "foo"

        s = fn.graph

        FileCheck().check("x : Union(int, str)").run(s)

    def test_union_redundant_arguments_are_skipped_optional(self):
        @torch.jit.script
        def fn(x: int | Optional[float] | Optional[int]) -> str:
            return "foo"

        s = fn.graph

        FileCheck().check("x : Union(float, int, NoneType)").run(s)

    def test_union_redundant_arguments_are_skipped_subtyping(self):
        @torch.jit.script
        def fn(x: str | Tuple[Optional[int], int] | Tuple[int, int]) -> str:
            return "foo"

        s = fn.graph

        FileCheck().check("x : Union((int?, int), str)").run(s)

    def test_union_redundant_arguments_are_skipped_container(self):
        @torch.jit.script
        def fn(x: List[str] | List[float] | List[str]) -> str:
            return "foo"

        s = fn.graph

        FileCheck().check("x : Union(float[], str[])").run(s)

    def test_union_argument_order_is_ignored(self):
        @torch.jit.script
        def fn1(x: int | str) -> str:
            return "foo"

        @torch.jit.script
        def fn2(x: str | int) -> str:
            return "foo"

        for s in (fn1.graph, fn2.graph):
            FileCheck().check("x : Union(int, str)").run(s)

    def test_union_argument_order_is_ignored_container(self):
        @torch.jit.script
        def fn1(x: List[str] | List[int]) -> str:
            return "foo"

        @torch.jit.script
        def fn2(x: List[int] | List[str]) -> str:
            return "foo"

        for s in (fn1.graph, fn2.graph):
            FileCheck().check("x : Union(int[], str[])").run(s)

    def test_union_T_None_is_equivalent_to_optional_T(self):
        @torch.jit.script
        def inner(x: int | None) -> int:
            if x is not None:
                return x
            else:
                return 5

        @torch.jit.script
        def fn1() -> int:
            a: Optional[int] = 5
            b: Optional[int] = None
            a_ = inner(a)
            b_ = inner(b)
            return a_ + b_

        self.assertEqual(fn1(), 10)

        @torch.jit.script
        def inner2(x: Optional[int]) -> int:
            if x is not None:
                return x
            else:
                return 5

        @torch.jit.script
        def fn2() -> int:
            a: int | None = 5
            b: int | None = None
            a_ = inner(a)
            b_ = inner(b)
            return a_ + b_

        self.assertEqual(fn2(), 10)

    @unittest.expectedFailure
    def test_union_optional_of_union_return(self):
        @torch.jit.script
        def fn() -> None | str | int:
            y: Optional[int | str] = "foo"
            return y

    @unittest.expectedFailure
    def test_union_optional_of_union_is_flattened(self):
        @torch.jit.script
        def fn(flag: int) -> str | int | None:
            y: int | str | None = "foo"
            if flag == 0:
                x: Optional[int | str] = y
            elif flag == 1:
                x: Optional[int | str] = 1
            else:
                x: Optional[int | str] = None
            return x

        # Can't use `checkScript` because it will flag the fact that
        # the original code has `Optional[Union[int, str]]` but the
        # saved/loaded code has `Union[int, NoneType, str]` (even
        # though this is exactly what we want)
        self.assertEqual(fn(0), "foo")
        self.assertEqual(fn(1), 1)
        self.assertEqual(fn(2), None)

        buffer = io.BytesIO()
        torch.jit.save(fn, buffer)
        buffer = io.BytesIO(buffer.getvalue())
        l = torch.jit.load(buffer)

        s = l.code

        FileCheck().check("Union[int, NoneType, str]").check(
            "Union[int, NoneType, str]"
        ).run(s)

    def test_union_subclasses_larger_union(self):
        def fn() -> int | str | torch.Tensor:
            x: int | str = "foo"
            return x

        self.checkScript(fn, ())

    # TODO: We would like to eventually support this. The issue is being
    # tracked at https://github.com/pytorch/pytorch/issues/58167
    def test_union_as_dict_key(self):
        def fn():
            x: Dict[int | str, str] = {}
            x["foo"] = "bar"
            x[1] = 2
            return x[1]

        with self.assertRaisesRegex(
            RuntimeError,
            "only int, float, complex, Tensor, device and string keys are supported",
        ):
            torch.jit.script(fn)

    def test_union_as_dict_value(self):
        def fn():
            x: Dict[str, int | str] = {}
            x["foo"] = "bar"
            x["baz"] = 2
            return x["baz"]

        self.checkScript(fn, ())

    def test_union_module_with_union_instance_variable(self):
        class M(torch.nn.Module):
            x: int | str

            def __init__(self, x: int | str):
                super().__init__()
                self.x: int | str = x

            def forward(self, y: int | str):
                self.x = y
                return self.x

        self.checkModule(
            M(
                2,
            ),
            (1,),
        )
        self.checkModule(M("bar"), ("foo",))

    def test_union_module_with_union_class_variable(self):
        class M(torch.nn.Module):
            x: int | str = "foo"

            def __init__(self, y: int):
                super().__init__()
                x = y

            def forward(self, z: str):
                x = z
                return x

        self.checkModule(M(1), ("foo",))

    def test_union_type_refinement(self):
        def fn(x: int | str) -> str:
            if isinstance(x, str):
                z = x + "bar"
                return x
            else:
                return "baz"

        self.checkScript(fn, ("foo",))
        self.checkScript(fn, (1,))

    def test_union_type_refinement_union_rhs(self):
        def fn(x: int) -> str:
            if torch.jit.isinstance(x, int | str):
                return "bar"
            else:
                return "baz"

        self.checkScript(fn, (1,))

    def test_union_type_refinement_tuple_rhs(self):
        def fn(x: int | float | List[str]) -> str:
            if isinstance(x, (int, float)):
                if isinstance(x, int):
                    return str(x)
                else:
                    return "foo"
            else:
                if len(x):
                    return x[0]
                else:
                    return "bar"

        self.checkScript(fn, (1,))
        self.checkScript(fn, (1.0,))
        self.checkScript(fn, (["a", "b", "c"],))

    def test_union_type_refinement_tuple_rhs_noncontained_type(self):
        def fn(x: int | List[str]) -> str:
            if isinstance(x, (int, float)):
                y = x + x
                return str(y)
            else:
                if len(x):
                    return x[0]
                else:
                    return "bar"

        self.checkScript(fn, (1,))
        self.checkScript(fn, (["a", "b", "c"],))

    def test_union_type_refinement_tuple_rhs_union(self):
        @torch.jit.script
        def fn(x: int) -> str:
            if torch.jit.isinstance(x, (int | str, float)):
                y = x + x
                return str(y)
            else:
                return "foo"

        # TODO: There's currently an unrelated bug in
        # `torch.jit.isinstance` that makes it fail for tuple literals.
        # Posted here: https://github.com/pytorch/pytorch/issues/60095
        # Change `assertEqual` to `checkScript` when the bug is fixed
        self.assertEqual(fn(1), "2")

    def test_union_type_refinement_statically_false(self):
        @torch.jit.script
        def fn(x: int) -> str:
            if torch.jit.isinstance(x, (str | float, List[str], str)):
                z = x + "foo"
                return z
            else:
                return "bar"

        s = fn.graph

        # Check that we don't have any branching statements
        FileCheck().check_not("block0()").check_not("block1()").run(s)

    def test_union_type_refinement_statically_true(self):
        @torch.jit.script
        def fn(x: List[int] | int) -> List[int] | int:
            if not torch.jit.isinstance(x, (int, List[int])):
                return x
            else:
                l = [1, 2, 3]
                y: List[int] | int = l
                return y

        s = fn.graph

        # Check that we don't have any branching statements
        FileCheck().check_not("block0()").check_not("block1()").run(s)

    def test_union_type_refinement_partial_static_refinement_tuple_rhs(self):
        def fn(x: List[int] | int) -> int:
            if torch.jit.isinstance(x, (int, float, str)):
                # We should know that `x` is an `int` here
                z = x + 1
                return z
            else:
                return 100

        self.checkScript(fn, ([1, 2, 3],))
        self.checkScript(fn, (1,))

    def test_union_type_refinement_partial_static_refinement_union_rhs(self):
        def fn(x: List[int] | int) -> int:
            if torch.jit.isinstance(x, int | float | str):
                # We should know that `x` is an `int` here
                z = x + 1
                return z
            else:
                return 100

        self.checkScript(fn, ([1, 2, 3],))
        self.checkScript(fn, (1,))

    def test_union_type_refinement_internal_declaration(self):
        def fn(flag: bool) -> str:
            x: int | str | None = None
            if flag:
                y = "foo"
            else:
                y = 1
            if isinstance(x, str):
                return x
            else:
                return "bar"

        self.checkScript(fn, (True,))
        self.checkScript(fn, (False,))

    def test_union_branching_with_union_return_and_homogenous_types(self):
        def fn(x: int) -> int | str:
            if x % 2:
                return "foo"
            else:
                return "bar"

        self.checkScript(fn, (1,))
        self.checkScript(fn, (8,))

    def test_union_branching_does_not_autoinfer_undeclared_union(self):
        def fn(x: int) -> str:
            if x % 2:
                y = "foo"
            else:
                y = x
            if isinstance(y, str):
                return y
            else:
                return "bar"

        with self.assertRaisesRegex(
            RuntimeError,
            "y is set to type str in the true branch and type int in the false branch",
        ):
            torch.jit.script(fn)

    def test_union_branching_does_not_widen_existing_inferred_type(self):
        def fn(x: int) -> str:
            y = "foo"
            if x % 2:
                y = "bar"
            else:
                y = x
            if isinstance(y, str):
                return y
            else:
                return "baz"

        with self.assertRaisesRegex(
            RuntimeError,
            "previously had type str but is now being assigned to a value of type int",
        ):
            torch.jit.script(fn)

    def test_union_schema_matching_on_internal_type(self):
        def fn(x: List[int] | Dict[str, int]) -> int:
            if torch.jit.isinstance(x, List[int]):
                return x[0]
            else:
                return list(x.values())[0]

        self.checkScript(fn, ([1, 2, 3],))
        self.checkScript(fn, ({"foo": 1, "bar": 2, "baz": 3},))

    def test_union_subtractive_refinement(self):
        def fn(x: List[int] | int) -> int:
            if not isinstance(x, int):
                x.append(1)
                return x[0]
            else:
                return x

        self.checkScript(fn, (1,))
        self.checkScript(fn, ([1, 2, 3],))

    def test_union_subtractive_refinement_with_container(self):
        def fn(x: List[int] | int) -> int:
            if not torch.jit.isinstance(x, List[int]):
                return x
            else:
                x.append(1)
                return x[0]

        self.checkScript(fn, (1,))
        self.checkScript(fn, ([1, 2, 3],))

    def test_union_memory_aliasing(self):
        def fn():
            x: List[torch.Tensor] = []
            z: List[Optional[List[torch.Tensor]]] = []
            z.append(x)
            x_alias = z[0]
            if torch.jit.isinstance(x_alias, List[torch.Tensor]):
                x_alias.append(torch.tensor(3))
            return x

        self.checkScript(fn, ())

    def test_union_serialization_preserves_type_annotations(self):
        # This function will fail after being torch.jit.save'd and
        # torch.jit.load'd if the type annotations aren't preserved
        # for Union during serialization. We need the `Union[str, int]`
        # annotation to make sure that `y` is typed as a Union instead
        # of as a str in one branch and an int in the other
        def fn(x: int) -> str:
            if x % 2:
                y: str | int = "bar"
            else:
                y: str | int = x
            if isinstance(y, str):
                return y
            else:
                return "baz"

        self.checkScript(fn, (1,))
        self.checkScript(fn, (8,))

    def _assert_passes(self, template: str, ann: str, lhs: str):
        code = template.format(ann=ann, lhs=lhs)
        self.checkScript(code, (), name="fn")

    def _assert_raises(self, template: str, ann: str, lhs: str, msg: str):
        code = template.format(ann=ann, lhs=lhs)
        with self.assertRaisesRegex(RuntimeError, msg):
            cu = torch.jit.CompilationUnit(code, _frames_up=1)
            string_frontend = getattr(cu, "fn")  # noqa: B009

    def test_union_with_list_assignment(self):
        template = dedent(
            """
            def fn():
                x: {ann} = {lhs}
                if torch.jit.isinstance(x, List[torch.Tensor]):
                    x.append(torch.tensor(3))
                return x
        """
        )

        lhs = {
            "list_literal_empty": "[]",
            "list_literal_of_tensor": "[torch.arange(3), torch.arange(5)]",
            "list_literal_of_str": '["foo", "bar", "baz"]',
            "list_literal_of_mixed": "[torch.arange(5), 1]",
            "list_comprehension_of_tensor": "[torch.add(x, 1) for x in [torch.arange(3), torch.arange(5)]]",
            "list_comprehension_of_str": '[x + "!" for x in ["foo", "bar", "baz"]]',
            "list_comprehension_of_mixed": "[torch.add(1, x) for x in [torch.arange(5), 1]]",
        }

        """
        List[str] | List[torch.Tensor]
        """
        self._assert_raises(
            template,
            "List[str] | List[torch.Tensor]",
            lhs["list_literal_empty"],
            "there are multiple possible List type candidates in the Union annotation",
        )

        self._assert_passes(
            template, "List[str] | List[torch.Tensor]", lhs["list_literal_of_tensor"]
        )

        self._assert_passes(
            template, "List[str] | List[torch.Tensor]", lhs["list_literal_of_str"]
        )

        self._assert_raises(
            template,
            "List[str] | List[torch.Tensor]",
            lhs["list_literal_of_mixed"],
            "none of those types match the types of the given list elements",
        )

        self._assert_passes(
            template,
            "List[str] | List[torch.Tensor]",
            lhs["list_comprehension_of_tensor"],
        )

        self._assert_passes(
            template, "List[str] | List[torch.Tensor]", lhs["list_comprehension_of_str"]
        )

        # TODO: Support mixed list comprehensions
        self._assert_raises(
            template,
            "List[str] | List[torch.Tensor]",
            lhs["list_comprehension_of_mixed"],
            "Arguments for call are not valid",
        )

        """
        int | torch.Tensor
        """
        self._assert_raises(
            template,
            "int | torch.Tensor",
            lhs["list_literal_empty"],
            "Expected an Union type annotation with an inner List type",
        )

        self._assert_raises(
            template,
            "int | torch.Tensor",
            lhs["list_literal_of_tensor"],
            "Expected an Union type annotation with an inner List type",
        )

        self._assert_raises(
            template,
            "int | torch.Tensor",
            lhs["list_comprehension_of_tensor"],
            "Expected an Union type annotation with an inner List type",
        )

        """
        List[torch.Tensor] | int
        """
        self._assert_passes(
            template, "List[torch.Tensor] | int", lhs["list_literal_empty"]
        )

        self._assert_passes(
            template, "List[torch.Tensor] | int", lhs["list_literal_of_tensor"]
        )

        self._assert_raises(
            template,
            "List[torch.Tensor] | int",
            lhs["list_literal_of_str"],
            r"List type annotation `List\[Tensor\]` did "
            "not match the types of the given list "
            "elements",
        )

        self._assert_raises(
            template,
            "List[torch.Tensor] | int",
            lhs["list_literal_of_mixed"],
            r"List type annotation `List\[Tensor\]` did "
            "not match the types of the given list "
            "elements",
        )

        self._assert_passes(
            template, "List[torch.Tensor] | int", lhs["list_comprehension_of_tensor"]
        )

        self._assert_raises(
            template,
            "List[torch.Tensor] | int",
            lhs["list_comprehension_of_str"],
            r"List type annotation `List\[Tensor\]` did "
            "not match the types of the given list "
            "elements",
        )

        # TODO(@ansley): Support mixed list comprehensions
        self._assert_raises(
            template,
            "List[torch.Tensor] | int",
            lhs["list_comprehension_of_mixed"],
            "Arguments for call are not valid",
        )

    def test_union_with_dict_assignment(self):
        template = dedent(
            """
            def fn():
                x: {ann} = {lhs}
                if torch.jit.isinstance(x, Dict[str, torch.Tensor]):
                    x["foo"] = torch.tensor(3)
                return x
        """
        )

        lhs = {
            "dict_literal_empty": "{}",
            "dict_literal_of_str_tensor": '{"foo" : torch.arange(3), "bar" : torch.arange(5)}',
            "dict_literal_of_str_int": '{"foo" : 1, "bar" : 2}',
            "dict_literal_of_mixed": '{"foo" : torch.arange(3), "bar" : 2}',
            "dict_comprehension_of_str_tensor": '{x : torch.add(y, 1) for x, y in \
                    zip(["foo", "bar"], [torch.arange(3), torch.arange(5)])}',
            "dict_comprehension_of_str_int": '{x : torch.add(y, 1) for x, y in \
                    zip(["foo", "bar"], [1, 2]}',
            "dict_comprehension_of_mixed": '{x : torch.add(y, 1) for x, y in \
                    zip(["foo", "bar"], [torch.arange(3), 2])}',
            "dict_keyword": "dict(foo=torch.arange(3), baz=torch.arange(5))",
            "dict_keyword_with_iterable": 'dict([("foo", torch.arange(3)), ("bar", torch.arange(5))])',
            "dict_keyword_with_empty_iterable": "dict([])",
            "dict_keyword_with_internal_aggregate_function": 'dict(zip(["foo", "bar"], [torch.arange(3), torch.arange(5)])',
            "dict_keyword_with_mapping": 'dict({"foo" : torch.arange(3), "bar" : torch.arange(5)})',
            "dict_keyword_with_mapping_and_kwargs": 'dict({"foo" : torch.arange(3), "bar" : torch.arange(5)}, baz=torch.arange(7))',
        }

        """
        Dict[str, torch.Tensor] | Dict[str, int]
        """
        self._assert_raises(
            template,
            "List[str] | List[torch.Tensor]",
            lhs["dict_literal_empty"],
            "Expected an Union type annotation with an inner Dict type",
        )

        self._assert_passes(
            template,
            "Dict[str, torch.Tensor] | Dict[str, int]",
            lhs["dict_literal_of_str_tensor"],
        )

        self._assert_passes(
            template,
            "Dict[str, torch.Tensor] | Dict[str, int]",
            lhs["dict_literal_of_str_int"],
        )

        self._assert_raises(
            template,
            "Dict[str, torch.Tensor] | Dict[str, int]",
            lhs["dict_literal_of_mixed"],
            "none of those dict types can hold the types of the given keys and values",
        )

        # TODO: String frontend does not support tuple unpacking
        # https://github.com/pytorch/pytorch/issues/64096
        # self._assert_passes(template, "Dict[str, torch.Tensor] | Dict[str, int]",
        #              lhs["dict_comprehension_of_str_tensor"])

        # self._assert_passes(template, "Dict[str, torch.Tensor] | Dict[str, int]",
        #              lhs["dict_comprehension_of_str_int"])

        # self._assert_raises(template, "Dict[str, torch.Tensor] | Dict[str, int]",
        #              lhs["dict_comprehension_of_mixed"],
        #              "foobar")

        # self._assert_passes(template,
        #                    "Dict[str, torch.Tensor] | Dict[str, int]",
        #                    lhs["dict_keyword_with_internal_aggregate_function"])

        # TODO(@ansley): Follow-up project needed for full type
        # inference with dict keyword (supported for dict comprehension
        # and dict literal already; should not be a blocker for anyone)
        self._assert_raises(
            template,
            "Dict[str, torch.Tensor] | Dict[str, int]",
            lhs["dict_keyword"],
            "full type inference is not yet supported",
        )

        self._assert_raises(
            template,
            "Dict[str, torch.Tensor] | Dict[str, int]",
            lhs["dict_keyword_with_iterable"],
            "full type inference is not yet supported",
        )

        self._assert_raises(
            template,
            "Dict[str, torch.Tensor] | Dict[str, int]",
            lhs["dict_keyword_with_empty_iterable"],
            "full type inference is not yet supported",
        )

        self._assert_raises(
            template,
            "Dict[str, torch.Tensor] | Dict[str, int]",
            lhs["dict_keyword_with_mapping"],
            "full type inference is not yet supported",
        )

        self._assert_raises(
            template,
            "Dict[str, torch.Tensor] | Dict[str, int]",
            lhs["dict_keyword_with_mapping_and_kwargs"],
            "full type inference is not yet supported",
        )

        """
        int | torch.Tensor
        """
        self._assert_raises(
            template,
            "int | torch.Tensor",
            lhs["dict_literal_empty"],
            "Expected an Union type annotation with an inner Dict type",
        )

        self._assert_raises(
            template,
            "int | torch.Tensor",
            lhs["dict_literal_of_str_tensor"],
            "Expected an Union type annotation with an inner Dict type",
        )

        # See above--string frontend does not support tuple unpacking
        # self._assert_raises(template, "int | torch.Tensor",
        #              lhs["dict_comprehension_of_tensor"],
        #              "foobar")

        """
        Dict[str, torch.Tensor] | int
        """
        self._assert_passes(
            template, "Dict[str, torch.Tensor] | int", lhs["dict_literal_empty"]
        )

        self._assert_passes(
            template, "Dict[str, torch.Tensor] | int", lhs["dict_literal_of_str_tensor"]
        )

        self._assert_raises(
            template,
            "Dict[str, torch.Tensor] | int",
            lhs["dict_literal_of_str_int"],
            "Type annotation was inferred to be "
            r"`Dict\[str, Tensor\]`, but the type of "
            "values given by the dict literal is",
        )

        self._assert_raises(
            template,
            "Dict[str, torch.Tensor] | int",
            lhs["dict_literal_of_mixed"],
            "Type annotation was inferred to be "
            r"`Dict\[str, Tensor\]`, but the type of "
            "values given by the dict literal is",
        )

        self._assert_passes(
            template, "Dict[str, torch.Tensor] | int", lhs["dict_keyword"]
        )

        self._assert_passes(
            template, "Dict[str, torch.Tensor] | int", lhs["dict_keyword_with_iterable"]
        )

        self._assert_passes(
            template,
            "Dict[str, torch.Tensor] | int",
            lhs["dict_keyword_with_empty_iterable"],
        )

        self._assert_passes(
            template, "Dict[str, torch.Tensor] | int", lhs["dict_keyword_with_mapping"]
        )

        self._assert_passes(
            template,
            "Dict[str, torch.Tensor] | int",
            lhs["dict_keyword_with_mapping_and_kwargs"],
        )

        # See above--string frontend does not support tuple unpacking
        # self._assert_passes(template,
        #                    "Dict[str, torch.Tensor] | int",
        #                    lhs["dict_keyword_with_internal_aggregate_function"])
        #
        # self._assert_passes(template,
        #                    "Dict[str, torch.Tensor] | int",
        #                    lhs["dict_comprehension_of_str_tensor"])

        # self._assert_raises(template,
        #                    "Dict[str, torch.Tensor] | int",
        #                    lhs["dict_comprehension_of_str_int"],
        #                    "foobar")

        # self._assert_raises(template,
        #                    "Dict[str, torch.Tensor] | int",
        #                    lhs["dict_comprehension_of_mixed"],
        #                    "foobar")


if __name__ == "__main__":
    raise_on_run_directly("test/test_jit.py")



#!/usr/bin/env python3
# Owner(s): ["oncall: r2p"]

import functools

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory  of this source tree.
import json
import os
import signal
import unittest
import uuid
from multiprocessing.pool import ThreadPool
from typing import Any
from unittest.mock import call, MagicMock, patch

import torch.distributed as dist
import torch.distributed.elastic.rendezvous.registry as rdzv_registry
from torch.distributed.elastic.agent.server.api import (
    _get_fq_hostname,
    _RoleInstanceInfo,
    RunResult,
    SimpleElasticAgent,
    Worker,
    WorkerGroup,
    WorkerSpec,
    WorkerState,
)
from torch.distributed.elastic.events import EventSource
from torch.distributed.elastic.multiprocessing import SignalException
from torch.distributed.elastic.multiprocessing.errors import ProcessFailure
from torch.distributed.elastic.rendezvous import RendezvousHandler, RendezvousParameters
from torch.distributed.elastic.rendezvous.api import RendezvousGracefulExitError
from torch.distributed.elastic.utils.distributed import get_free_port


def do_nothing():
    pass


class WorkerStateTest(unittest.TestCase):
    def test_is_running(self):
        for state in WorkerState:
            if state == WorkerState.HEALTHY or state == WorkerState.UNHEALTHY:
                self.assertTrue(WorkerState.is_running(state))
            else:
                self.assertFalse(WorkerState.is_running(state))


class WorkerGroupTest(unittest.TestCase):
    def test_worker_group_constructor(self):
        spec = WorkerSpec(
            role="test_trainer",
            local_world_size=4,
            fn=do_nothing,
            args=(),
            rdzv_handler=None,
            max_restarts=50,
            monitor_interval=0.1,
        )
        worker_group = WorkerGroup(spec)

        self.assertEqual(WorkerState.INIT, worker_group.state)

        workers = worker_group.workers
        self.assertEqual(4, len(workers))

        # validate full, consecutive local ranks
        self.assertSetEqual(set(range(4)), {w.local_rank for w in workers})

        # global_rank, world_size are assigned after rdzv
        # id is assigned after starting worker (by the agent)
        # validate there are None
        for w in workers:
            self.assertEqual(-1, w.global_rank)
            self.assertEqual(-1, w.world_size)
            self.assertEqual(None, w.id)

        # rank and store are assigned after rdzv; validate that they are None
        self.assertIsNone(worker_group.group_rank)
        self.assertIsNone(worker_group.store)


class RoleInstanceInfoTest(unittest.TestCase):
    def test_compare(self):
        agent_role1 = _RoleInstanceInfo("role", 1, 10)
        agent_role2 = _RoleInstanceInfo("role", 2, 10)
        self.assertEqual(1, _RoleInstanceInfo.compare(agent_role2, agent_role1))
        agent_role1 = _RoleInstanceInfo("role1", 1, 10)
        agent_role2 = _RoleInstanceInfo("role2", 2, 10)
        self.assertEqual(-1, _RoleInstanceInfo.compare(agent_role1, agent_role2))
        agent_role1 = _RoleInstanceInfo("role1", 1, 10)
        agent_role2 = _RoleInstanceInfo("role2", 1, 10)
        self.assertEqual(-1, _RoleInstanceInfo.compare(agent_role1, agent_role2))

    def test_serde(self):
        agent_role = _RoleInstanceInfo("role", 1, 10)
        str_data = agent_role.serialize()
        actual_agent_role = _RoleInstanceInfo.deserialize(str_data)
        self.assertEqual(agent_role.role, actual_agent_role.role)
        self.assertEqual(agent_role.rank, actual_agent_role.rank)
        self.assertEqual(
            agent_role.local_world_size, actual_agent_role.local_world_size
        )

    def test_find_boundaries(self):
        role_infos = [
            _RoleInstanceInfo("trainer", 1, 1),
            _RoleInstanceInfo("trainer", 2, 2),
            _RoleInstanceInfo("trainer", 3, 3),
            _RoleInstanceInfo("parameter_server", 4, 5),
            _RoleInstanceInfo("parameter_server", 0, 4),
        ]
        start_idx, end_idx = _RoleInstanceInfo.find_role_boundaries(
            role_infos, "trainer"
        )
        self.assertEqual(start_idx, 0)
        self.assertEqual(end_idx, 2)


class TestAgent(SimpleElasticAgent):
    def __init__(self, spec):
        super().__init__(spec)
        self.stop_workers_call_count = 0
        self.start_workers_call_count = 0

    def _stop_workers(self, worker_group: WorkerGroup) -> None:
        # workers are fake, nothing to stop; just clear the rdzv info
        worker_group.group_rank = None
        worker_group.group_world_size = None
        self.stop_workers_call_count += 1

    def _start_workers(self, worker_group: WorkerGroup) -> dict[int, Any]:
        # crate fake workers; make worker id equal to global rank
        ids = {}
        for worker in worker_group.workers:
            ids[worker.local_rank] = worker.global_rank
        self.start_workers_call_count += 1
        return ids

    def _monitor_workers(self, worker_group: WorkerGroup) -> RunResult:
        raise NotImplementedError("mock this method")

    def _shutdown(self):
        pass


def monres(state: WorkerState):
    if state == WorkerState.SUCCEEDED:
        return RunResult(state=state, return_values={0: 0}, failures={})
    elif state in {WorkerState.UNHEALTHY, WorkerState.FAILED}:
        pf = ProcessFailure(local_rank=0, pid=999, exitcode=1, error_file="<none>")
        return RunResult(state=state, return_values={}, failures={0: pf})
    else:
        return RunResult(state=state)


class RecordWorkerEventsTest(unittest.TestCase):
    def setUp(self):
        self.spec = MagicMock()
        self.spec.role = "test_role"
        self.spec.get_entrypoint_name.return_value = "test_entrypoint"
        self.spec.rdzv_handler.get_run_id.return_value = "test_run_id"
        self.spec.rdzv_handler.get_backend.return_value = "test_backend"
        self.spec.max_restarts = 3

        self.agent = TestAgent(self.spec)

        # Create a mock worker spec and agent
        self.agent._worker_group = MagicMock()
        self.agent._worker_group.spec = MagicMock()
        self.agent._worker_group.spec.event_log_handler = "test_handler"

        # Setup worker group
        self.worker_group = WorkerGroup(self.spec)
        self.worker_group.group_world_size = 2
        self.worker_group.group_rank = 1
        self.agent._worker_group = self.worker_group

        # Create a test worker

        self.workers = [
            Worker(
                local_rank=0,
                global_rank=0,
                role_rank=0,
                world_size=2,
                role_world_size=2,
            ),
            Worker(
                local_rank=1,
                global_rank=1,
                role_rank=1,
                world_size=2,
                role_world_size=2,
            ),
        ]
        self.workers[0].id = 0
        self.workers[1].id = 1
        self.agent._worker_group.workers = self.workers

    @patch("torch.distributed.elastic.agent.server.api.record")
    def test_record_worker_events_success(self, mock_record):
        # Create a RunResult with successful workers
        result = RunResult(
            state=WorkerState.SUCCEEDED,
            return_values={0: "result0", 1: "result1"},
            failures={},
        )

        # Call the method under test
        self.agent._record_worker_events(result)

        # Verify record was called twice (once for each worker)
        self.assertEqual(mock_record.call_count, 2)

        # Check that both calls were for SUCCEEDED events
        for call_args in mock_record.call_args_list:
            event = call_args[0][0]

            self.assertEqual(event.source, EventSource.WORKER)
            self.assertEqual(event.metadata["state"], "SUCCEEDED")
            self.assertIsNone(event.metadata["raw_error"])
            md = json.loads(event.metadata["metadata"])
            self.assertEqual(md["exit_code"], [None])
            self.assertEqual(md["worker_pid"], [None])

    @patch("torch.distributed.elastic.agent.server.api.record")
    def test_record_worker_events_failure(self, mock_record):
        # Create failures with error data
        failure0 = ProcessFailure(
            local_rank=0, pid=1000, exitcode=1, error_file="error0.json"
        )

        # Create a RunResult with one failed worker and one terminated worker
        result = RunResult(
            state=WorkerState.FAILED,
            return_values={},
            failures={0: failure0},  # Only worker 0 has a specific failure
        )

        # Call the method under test
        self.agent._record_worker_events(result)

        # Verify record was called twice (once for each worker)
        self.assertEqual(mock_record.call_count, 2)

        # Get the calls
        calls = mock_record.call_args_list

        # Check first call for the failed worker (global_rank=0)
        failed_event = calls[0][0][0]
        self.assertEqual(failed_event.source, EventSource.WORKER)
        self.assertEqual(failed_event.metadata["state"], "FAILED")
        self.assertEqual(failed_event.metadata["global_rank"], 0)
        md = json.loads(failed_event.metadata["metadata"])
        self.assertEqual(
            failed_event.metadata["raw_error"],
            '{"message": "<NONE>", "errorTraits": {"category": "system_terminated_error", "retryability": "False"}}',
        )
        self.assertEqual(md["exit_code"], [1])
        self.assertEqual(md["worker_pid"], [1000])

        # Check second call for the terminated worker (global_rank=1)
        terminated_event = calls[1][0][0]
        self.assertEqual(terminated_event.source, EventSource.WORKER)
        self.assertEqual(terminated_event.metadata["state"], "TERMINATED")
        self.assertEqual(terminated_event.metadata["global_rank"], 1)
        self.assertIsNone(terminated_event.metadata["raw_error"])
        md = json.loads(terminated_event.metadata["metadata"])
        self.assertEqual(md["exit_code"], [None])
        self.assertEqual(md["worker_pid"], [None])


class ConstructEventTest(unittest.TestCase):
    def setUp(self):
        # Create minimal spec and agent for testing
        self.spec = MagicMock()
        self.spec.role = "test_role"
        self.spec.get_entrypoint_name.return_value = "test_entrypoint"
        self.spec.rdzv_handler.get_run_id.return_value = "test_run_id"
        self.spec.rdzv_handler.get_backend.return_value = "test_backend"
        self.spec.max_restarts = 3

        self.agent = TestAgent(self.spec)
        self.agent._remaining_restarts = 2
        self.agent._total_execution_time = 42

        # Setup worker group
        self.worker_group = WorkerGroup(self.spec)
        self.worker_group.group_world_size = 2
        self.worker_group.group_rank = 1
        self.agent._worker_group = self.worker_group

        # Create a test worker
        self.worker = Worker(
            local_rank=0, global_rank=5, role_rank=3, world_size=8, role_world_size=4
        )
        self.worker.id = 12345

    def test_construct_event_agent_success(self):
        # Test constructing an agent success event
        event = self.agent._construct_event(state="SUCCEEDED", source=EventSource.AGENT)

        # Verify basic event properties
        self.assertEqual(event.name, "torchelastic.worker.status.SUCCEEDED")
        self.assertEqual(event.source, EventSource.AGENT)

        # Verify metadata
        metadata = event.metadata
        self.assertEqual(metadata["run_id"], "test_run_id")
        self.assertIsNone(metadata["global_rank"])
        self.assertEqual(metadata["group_rank"], 1)
        self.assertIsNone(metadata["worker_id"])
        self.assertEqual(metadata["role"], "test_role")
        self.assertEqual(metadata["state"], "SUCCEEDED")
        self.assertEqual(metadata["total_run_time"], 42)
        self.assertEqual(metadata["rdzv_backend"], "test_backend")
        self.assertIsNone(metadata["raw_error"])
        self.assertEqual(
            metadata["agent_restarts"], 1
        )  # max_restarts - remaining_restarts
        self.assertIsNone(metadata["duration_ms"])

        # Verify JSON metadata
        md_dict = json.loads(metadata["metadata"])
        self.assertEqual(md_dict["group_world_size"], 2)
        self.assertEqual(md_dict["entry_point"], "test_entrypoint")

    def test_construct_event_worker_failure(self):
        # Test constructing a worker failure event with raw error
        raw_error = json.dumps(
            {"error_message": "Test error", "traceback": "stack trace"}
        )
        event = self.agent._construct_event(
            state="FAILED",
            source=EventSource.WORKER,
            worker=self.worker,
            raw_error=raw_error,
            exit_code=1,
        )

        # Verify basic event properties
        self.assertEqual(event.name, "torchelastic.worker.status.FAILED")
        self.assertEqual(event.source, EventSource.WORKER)

        # Verify metadata
        metadata = event.metadata
        self.assertEqual(metadata["run_id"], "test_run_id")
        self.assertEqual(metadata["global_rank"], 5)
        self.assertEqual(metadata["group_rank"], 1)
        self.assertEqual(metadata["worker_id"], "12345")
        self.assertEqual(metadata["role"], "test_role")
        self.assertEqual(metadata["state"], "FAILED")
        self.assertEqual(metadata["total_run_time"], 42)
        self.assertEqual(metadata["rdzv_backend"], "test_backend")
        self.assertEqual(metadata["raw_error"], raw_error)
        self.assertEqual(metadata["agent_restarts"], 1)

        # Verify worker-specific metadata
        md_dict = json.loads(metadata["metadata"])
        self.assertEqual(md_dict["local_rank"], [0])
        self.assertEqual(md_dict["role_rank"], [3])
        self.assertEqual(md_dict["role_world_size"], [4])
        self.assertEqual(md_dict["exit_code"], [1])

    def test_construct_event_with_duration(self):
        # Test constructing an event with duration_ms
        event = self.agent._construct_event(
            state="RENDEZVOUS", source=EventSource.AGENT, duration_ms=123.45
        )

        # Verify duration is set correctly
        self.assertEqual(event.metadata["duration_ms"], 123.45)

    def test_construct_event_worker_no_error(self):
        # Test constructing a worker event without error info
        event = self.agent._construct_event(
            state="HEALTHY", source=EventSource.WORKER, worker=self.worker
        )

        # Verify error fields are None
        metadata = event.metadata
        self.assertIsNone(metadata["raw_error"])

        # Check worker info is set
        self.assertEqual(metadata["global_rank"], 5)
        self.assertEqual(metadata["worker_id"], "12345")

        # Check metadata JSON
        md_dict = json.loads(metadata["metadata"])
        self.assertEqual(md_dict["local_rank"], [0])
        self.assertEqual(md_dict["role_rank"], [3])
        self.assertEqual(md_dict["role_world_size"], [4])
        self.assertNotIn("exit_code", [None])


class SimpleElasticAgentTest(unittest.TestCase):
    def _get_worker_spec(
        self,
        max_restarts=1,
        monitor_interval=0.1,
        role="test_trainer",
        local_world_size=8,
        local_addr=None,
        event_log_handler="null",
    ):
        run_id = str(uuid.uuid4().int)
        port = get_free_port()
        if local_addr is None:
            endpoint = f"127.0.0.1:{port}"
        else:
            endpoint = f"{local_addr}:{port}"

        rdzv_params = RendezvousParameters(
            backend="static",
            endpoint=endpoint,
            run_id=run_id,
            min_nodes=1,
            max_nodes=1,
            rank=0,
        )
        rdzv_handler = rdzv_registry.get_rendezvous_handler(rdzv_params)
        spec = WorkerSpec(
            role=role,
            local_world_size=local_world_size,
            fn=do_nothing,
            args=(),
            rdzv_handler=rdzv_handler,
            max_restarts=max_restarts,
            monitor_interval=monitor_interval,
            local_addr=local_addr,
            event_log_handler=event_log_handler,
        )
        return spec

    def test_agent_constructor(self):
        spec = self._get_worker_spec(max_restarts=1)
        agent = TestAgent(spec)
        worker_group = agent.get_worker_group()
        self.assertEqual(WorkerState.INIT, worker_group.state)
        self.assertEqual(spec.max_restarts, agent._remaining_restarts)

    @patch("torch.distributed.elastic.agent.server.api.put_metric")
    def test_record_flakiness_metric(self, put_metric_mock):
        spec = self._get_worker_spec(max_restarts=1)
        agent = TestAgent(spec)
        agent._record_flakiness_metric()
        put_metric_mock.assert_called_with("workers.test_trainer.flakiness", 0)
        agent._worker_group.spec.max_restarts = 10
        agent._remaining_restarts = 3
        agent._record_flakiness_metric()
        put_metric_mock.assert_called_with("workers.test_trainer.flakiness", 63)

    @patch("torch.distributed.elastic.agent.server.api.put_metric")
    def test_record_flakiness_metric_zero_restarts(self, put_metric_mock):
        spec = self._get_worker_spec(max_restarts=1)
        spec.max_restarts = 0
        agent = TestAgent(spec)
        agent._record_flakiness_metric()
        put_metric_mock.assert_called_with("workers.test_trainer.flakiness", 0)

    @patch("torch.distributed.elastic.agent.server.api.put_metric")
    def test_record_flakiness_metric_user_exception(self, put_metric_mock):
        spec = self._get_worker_spec(max_restarts=1)
        agent = TestAgent(spec)
        agent._record_flakiness_metric(True)
        put_metric_mock.assert_called_with("workers.test_trainer.flakiness", 100)

    @patch.object(TestAgent, "_invoke_run")
    @patch.object(TestAgent, "_record_metrics")
    @patch.object(TestAgent, "_record_worker_events")
    @patch.object(TestAgent, "_shutdown")
    def test_invoke_run(
        self, shutdown_mock, record_events_mock, record_metrics_mock, invoke_run_mock
    ):
        spec = self._get_worker_spec(max_restarts=1)
        agent = TestAgent(spec)
        agent.run()
        invoke_run_mock.assert_called_once()
        record_metrics_mock.assert_called_once()
        record_events_mock.assert_called_once()
        shutdown_mock.assert_called_once()

    def test_exit_barrier_sets_and_clears_flag(self):
        """Verify _in_exit_barrier is set before barrier and cleared after."""
        spec = self._get_worker_spec(max_restarts=0)
        agent = TestAgent(spec)
        agent._worker_group.state = WorkerState.SUCCEEDED
        agent._worker_group.group_world_size = 1
        agent._store = MagicMock()

        flag_during_barrier = []

        def mock_barrier(**kwargs):
            flag_during_barrier.append(agent._in_exit_barrier)

        with patch(
            "torch.distributed.elastic.utils.store.barrier",
            side_effect=mock_barrier,
        ):
            agent._exit_barrier()

        # Flag was True during barrier call
        self.assertTrue(flag_during_barrier[0])
        # Flag is False after barrier completes
        self.assertFalse(agent._in_exit_barrier)

    def test_exit_barrier_clears_flag_on_timeout(self):
        """Verify _in_exit_barrier is cleared even if barrier times out."""
        spec = self._get_worker_spec(max_restarts=0)
        agent = TestAgent(spec)
        agent._worker_group.state = WorkerState.SUCCEEDED
        agent._worker_group.group_world_size = 1
        agent._store = MagicMock()

        with patch(
            "torch.distributed.elastic.utils.store.barrier",
            side_effect=Exception("wait timeout"),
        ):
            agent._exit_barrier()

        # Flag must be cleared even on timeout
        self.assertFalse(agent._in_exit_barrier)

    @patch("torch.distributed.elastic.agent.server.api.put_metric")
    def test_record_metrics_success_no_retries(self, put_metric_mock):
        spec = self._get_worker_spec(max_restarts=1)
        agent = TestAgent(spec)
        group_result = RunResult({}, {})
        agent._record_metrics(group_result)
        calls = self._get_record_metrics_test_calls(success_no_retries=1)
        put_metric_mock.assert_has_calls(calls, any_order=True)

    @patch("torch.distributed.elastic.agent.server.api.put_metric")
    def test_record_metrics_success_with_retries(self, put_metric_mock):
        spec = self._get_worker_spec(max_restarts=10)
        agent = TestAgent(spec)
        agent._remaining_restarts = 2
        group_result = RunResult({}, {})
        agent._record_metrics(group_result)
        calls = self._get_record_metrics_test_calls(success_with_retries=1)
        put_metric_mock.assert_has_calls(calls, any_order=True)

    @patch("torch.distributed.elastic.agent.server.api.put_metric")
    def test_record_metrics_failed_with_retries(self, put_metric_mock):
        spec = self._get_worker_spec(max_restarts=10)
        agent = TestAgent(spec)
        agent._remaining_restarts = 2
        group_result = RunResult(
            state=WorkerState.FAILED, return_values={}, failures={0: 0}
        )
        agent._record_metrics(group_result)
        calls = self._get_record_metrics_test_calls(failed_with_retries=1)
        put_metric_mock.assert_has_calls(calls, any_order=True)

    @patch("torch.distributed.elastic.agent.server.api.put_metric")
    def test_record_metrics_failed_no_retries(self, put_metric_mock):
        spec = self._get_worker_spec(max_restarts=10)
        agent = TestAgent(spec)
        group_result = RunResult(
            state=WorkerState.FAILED, return_values={}, failures={0: 0}
        )
        agent._record_metrics(group_result)
        calls = self._get_record_metrics_test_calls(failed_no_retries=1)
        put_metric_mock.assert_has_calls(calls, any_order=True)

    def _get_record_metrics_test_calls(
        self,
        success_with_retries=0,
        success_no_retries=0,
        failed_with_retries=0,
        failed_no_retries=0,
    ):
        calls = [
            call("workers.test_trainer.run_success_with_retries", success_with_retries),
            call("workers.test_trainer.run_success_no_retries", success_no_retries),
            call("workers.test_trainer.run_failed_with_retries", failed_with_retries),
            call("workers.test_trainer.run_failed_no_retries", failed_no_retries),
        ]
        return calls

    def test_rendezvous(self):
        hostname = _get_fq_hostname()
        spec = self._get_worker_spec(max_restarts=1, local_addr=hostname)
        agent = TestAgent(spec)
        worker_group = agent.get_worker_group()
        agent._rendezvous(worker_group)

        # single agent rdzv
        self.assertEqual(1, worker_group.group_world_size)
        self.assertEqual(0, worker_group.group_rank)

        self.assertEqual(hostname, worker_group.master_addr)
        self.assertTrue(worker_group.master_port > 0)

        rank_set = {w.global_rank for w in worker_group.workers}
        for w in worker_group.workers:
            self.assertIsNone(w.id)
            local_world_size = spec.local_world_size
            group_world_size = worker_group.group_world_size
            group_rank = worker_group.group_rank

            self.assertEqual(local_world_size * group_world_size, w.world_size)
            self.assertEqual(
                local_world_size * group_rank + w.local_rank, w.global_rank
            )
            self.assertSetEqual(set(range(w.world_size)), rank_set)

    def test_rendezvous_default_master_addr(self):
        hostname = _get_fq_hostname()
        spec = self._get_worker_spec(max_restarts=1, local_addr=hostname)
        agent = TestAgent(spec)
        worker_group = agent.get_worker_group()
        agent._rendezvous(worker_group)

        self.assertEqual(_get_fq_hostname(), worker_group.master_addr)
        self.assertGreater(worker_group.master_port, 0)

    def test_rendezvous_master_addr_with_local_addr(self):
        spec_local_addr = "127.0.0.1"
        spec = self._get_worker_spec(max_restarts=1, local_addr=spec_local_addr)
        agent = TestAgent(spec)
        worker_group = agent.get_worker_group()
        agent._rendezvous(worker_group)

        self.assertNotEqual(_get_fq_hostname(), worker_group.master_addr)
        self.assertEqual(spec_local_addr, worker_group.master_addr)
        self.assertGreater(worker_group.master_port, 0)

    @patch.object(TestAgent, "_construct_event")
    @patch("torch.distributed.elastic.agent.server.api.record")
    def test_initialize_workers(self, mock_record, mock_construct_event):
        spec = self._get_worker_spec(max_restarts=1)
        agent = TestAgent(spec)
        worker_group = agent.get_worker_group()
        agent._initialize_workers(worker_group)

        self.assertEqual(WorkerState.HEALTHY, worker_group.state)
        for i in range(spec.local_world_size):
            worker = worker_group.workers[i]
            self.assertEqual(worker.id, worker.global_rank)

        mock_construct_event.assert_called()
        self.assertEqual(mock_construct_event.call_count, 10)
        mock_record.assert_called()
        second_arg = mock_record.call_args_list[0][0][1]
        self.assertEqual(second_arg, "null")

    @patch.object(TestAgent, "_construct_event")
    @patch("torch.distributed.elastic.agent.server.api.record")
    def test_initialize_workers_with_new_spec(self, mock_record, mock_construct_event):
        spec = self._get_worker_spec(
            max_restarts=1, event_log_handler="framework_logger"
        )
        agent = TestAgent(spec)
        worker_group = agent.get_worker_group()
        agent._initialize_workers(worker_group)

        self.assertEqual(WorkerState.HEALTHY, worker_group.state)
        for i in range(spec.local_world_size):
            worker = worker_group.workers[i]
            self.assertEqual(worker.id, worker.global_rank)

        mock_construct_event.assert_called()
        self.assertEqual(mock_construct_event.call_count, 10)
        mock_record.assert_called()
        second_arg = mock_record.call_args_list[0][0][1]
        self.assertEqual(second_arg, "framework_logger")

    def test_restart_workers(self):
        spec = self._get_worker_spec()
        agent = TestAgent(spec)
        worker_group = agent.get_worker_group()

        num_restarts = 3
        for _ in range(num_restarts):
            agent._restart_workers(worker_group)
            self.assertEqual(WorkerState.HEALTHY, worker_group.state)

            # test_rendezvous and test_initialize_workers
            # already validates the correctness of these fields
            # simply validate that they are not None
            # (e.g. that they get assigned)
            self.assertIsNotNone(worker_group.group_rank)
            self.assertIsNotNone(worker_group.group_world_size)
            for w in worker_group.workers:
                self.assertIsNotNone(w.id)
                self.assertIsNotNone(w.global_rank)
                self.assertIsNotNone(w.world_size)

        self.assertEqual(num_restarts, agent.start_workers_call_count)
        self.assertEqual(num_restarts, agent.stop_workers_call_count)

    @patch.object(
        TestAgent,
        "_monitor_workers",
        side_effect=[
            monres(WorkerState.HEALTHY),
            monres(WorkerState.HEALTHY),
            monres(WorkerState.SUCCEEDED),
        ],
    )
    @patch.object(TestAgent, "_record_worker_events")
    def test_run_happy_path(self, record_events_mock, mock_monitor_workers):
        # worker starts
        # is always healthy
        # then succeeds
        max_restarts = 10
        spec = self._get_worker_spec(max_restarts)
        agent = TestAgent(spec)

        agent.run()

        # no failure, no membership changes -> no retries
        self.assertEqual(max_restarts, agent._remaining_restarts)
        record_events_mock.assert_called_once()

    @patch.object(TestAgent, "_initialize_workers", side_effect=RuntimeError())
    def test_run_initialization_failure(self, mock_initialize_workers):
        spec = self._get_worker_spec()
        agent = TestAgent(spec)
        worker_group = agent._worker_group

        with self.assertRaises(RuntimeError):
            agent.run()

        self.assertEqual(WorkerState.INIT, worker_group.state)

    def test_run_max_retries_exceeded(self):
        for restartable_state in [
            monres(WorkerState.FAILED),
            monres(WorkerState.UNHEALTHY),
        ]:
            with patch.object(
                TestAgent, "_monitor_workers", return_value=restartable_state
            ) as mock_monitor_workers:
                spec = self._get_worker_spec(max_restarts=3, monitor_interval=0.1)
                agent = TestAgent(spec)
                worker_group = agent._worker_group

                agent.run()
                self.assertEqual(WorkerState.FAILED, worker_group.state)
                self.assertEqual(0, agent._remaining_restarts)
                # one monitor call for each retry + one to monitor the last retry
                self.assertEqual(spec.max_restarts + 1, mock_monitor_workers.call_count)

    @patch.object(
        TestAgent,
        "_monitor_workers",
        side_effect=[
            monres(WorkerState.HEALTHY),
            monres(WorkerState.HEALTHY),
            monres(WorkerState.HEALTHY),
            monres(WorkerState.SUCCEEDED),
        ],
    )
    @patch.object(RendezvousHandler, "num_nodes_waiting", side_effect=[1, 1, 0])
    @patch.object(TestAgent, "_record_worker_events")
    def test_run_membership_change(
        self, record_events_mock, mock_num_nodes_waiting, mock_monitor_workers
    ):
        spec = self._get_worker_spec(max_restarts=1, monitor_interval=0.1)
        agent = TestAgent(spec)
        worker_group = agent._worker_group

        agent.run()
        self.assertEqual(WorkerState.SUCCEEDED, worker_group.state)
        record_events_mock.assert_called_once()

    @patch.object(
        TestAgent, "_monitor_workers", return_value=monres(WorkerState.UNKNOWN)
    )
    def test_run_unknown_state(self, mock_monitor_workers):
        # when the state is unknown we exit immediately; no retries
        spec = self._get_worker_spec(max_restarts=100, monitor_interval=0.1)
        agent = TestAgent(spec)
        worker_group = agent._worker_group

        with self.assertRaises(Exception):
            agent.run()

        self.assertEqual(WorkerState.UNKNOWN, worker_group.state)
        self.assertEqual(1, mock_monitor_workers.call_count)
        self.assertEqual(spec.max_restarts, agent._remaining_restarts)

    def get_worker_assigned(self, store, role_infos_len, info) -> list[Worker]:
        i, role_info = info
        spec = self._get_worker_spec(
            max_restarts=3,
            monitor_interval=0.1,
            role=role_info.role,
            local_world_size=role_info.local_world_size,
        )
        agent = TestAgent(spec)
        workers = agent._assign_worker_ranks(
            store, role_info.rank, role_infos_len, spec
        )
        return [
            (
                w.local_rank,
                w.role_rank,
                w.global_rank,
                w.world_size,
                w.role_world_size,
            )
            for w in workers
        ]

    def test_assign_worker_ranks(self):
        role_infos = [
            _RoleInstanceInfo("parameter_server", 0, 4),
            _RoleInstanceInfo("trainer", 1, 1),
            _RoleInstanceInfo("trainer", 2, 2),
            _RoleInstanceInfo("trainer", 3, 3),
            _RoleInstanceInfo("parameter_server", 4, 5),
        ]
        store = dist.HashStore()

        f = functools.partial(self.get_worker_assigned, store, len(role_infos))

        with ThreadPool(len(role_infos)) as pool:
            out = pool.map(f, enumerate(role_infos))

        self.assertListEqual(
            out,
            [
                [
                    (0, 0, 0, 15, 9),
                    (1, 1, 1, 15, 9),
                    (2, 2, 2, 15, 9),
                    (3, 3, 3, 15, 9),
                ],
                [
                    (0, 0, 4, 15, 6),
                ],
                [
                    (0, 1, 5, 15, 6),
                    (1, 2, 6, 15, 6),
                ],
                [
                    (0, 3, 7, 15, 6),
                    (1, 4, 8, 15, 6),
                    (2, 5, 9, 15, 6),
                ],
                [
                    (0, 4, 10, 15, 9),
                    (1, 5, 11, 15, 9),
                    (2, 6, 12, 15, 9),
                    (3, 7, 13, 15, 9),
                    (4, 8, 14, 15, 9),
                ],
            ],
        )

    def test_assign_worker_ranks_indentical(self):
        os.environ["TORCH_ELASTIC_WORKER_IDENTICAL"] = "1"
        role_infos = [
            _RoleInstanceInfo("trainer", 0, 4),
            _RoleInstanceInfo("trainer", 1, 4),
            _RoleInstanceInfo("trainer", 2, 4),
            _RoleInstanceInfo("trainer", 3, 4),
            _RoleInstanceInfo("trainer", 4, 4),
        ]
        store = dist.HashStore()

        f = functools.partial(self.get_worker_assigned, store, len(role_infos))

        with ThreadPool(len(role_infos)) as pool:
            out = pool.map(f, enumerate(role_infos))

        self.assertListEqual(
            out,
            [
                [
                    (0, 0, 0, 20, 20),
                    (1, 1, 1, 20, 20),
                    (2, 2, 2, 20, 20),
                    (3, 3, 3, 20, 20),
                ],
                [
                    (0, 4, 4, 20, 20),
                    (1, 5, 5, 20, 20),
                    (2, 6, 6, 20, 20),
                    (3, 7, 7, 20, 20),
                ],
                [
                    (0, 8, 8, 20, 20),
                    (1, 9, 9, 20, 20),
                    (2, 10, 10, 20, 20),
                    (3, 11, 11, 20, 20),
                ],
                [
                    (0, 12, 12, 20, 20),
                    (1, 13, 13, 20, 20),
                    (2, 14, 14, 20, 20),
                    (3, 15, 15, 20, 20),
                ],
                [
                    (0, 16, 16, 20, 20),
                    (1, 17, 17, 20, 20),
                    (2, 18, 18, 20, 20),
                    (3, 19, 19, 20, 20),
                ],
            ],
        )
        os.environ["TORCH_ELASTIC_WORKER_IDENTICAL"] = "0"

    def test_get_event(self):
        spec = self._get_worker_spec(max_restarts=1)
        agent = TestAgent(spec)
        event = agent.get_event_succeeded()
        self.assertEqual("AGENT", event.source)
        self.assertEqual("static", event.metadata["rdzv_backend"])
        self.assertEqual("SUCCEEDED", event.metadata["state"])
        self.assertEqual(spec.role, event.metadata["role"])

    def test_get_worker_status_event(self):
        spec = self._get_worker_spec(max_restarts=4)
        agent = TestAgent(spec)
        agent._remaining_restarts = spec.max_restarts - 2
        actual_event = agent._construct_event(
            state="SUCCEEDED",
            source="WORKER",
            worker=agent._worker_group.workers[0],
        )
        self.assertEqual("WORKER", actual_event.source)
        self.assertEqual("static", actual_event.metadata["rdzv_backend"])
        self.assertEqual("SUCCEEDED", actual_event.metadata["state"])
        self.assertEqual(spec.role, actual_event.metadata["role"])
        self.assertEqual(2, actual_event.metadata["agent_restarts"])

    @patch("torch.distributed.elastic.agent.server.api.put_metric")
    @patch.object(TestAgent, "_invoke_run")
    def test_agent_process_signal_exception(self, invoke_run, _):
        spec = self._get_worker_spec(max_restarts=0)
        agent = TestAgent(spec)
        invoke_run.side_effect = SignalException(
            "signal exception", sigval=signal.SIGTERM
        )
        with patch.object(agent, "_shutdown") as shutdown_mock:
            with self.assertRaises(SignalException):
                agent.run()
            args, _ = shutdown_mock.call_args
            self.assertEqual(signal.SIGTERM, args[0])

    @patch("torch.distributed.elastic.agent.server.api.put_metric")
    @patch.object(TestAgent, "_invoke_run")
    def test_agent_process_handler_graceful_exception(self, invoke_run, _):
        spec = self._get_worker_spec(max_restarts=0)
        agent = TestAgent(spec)
        invoke_run.side_effect = RendezvousGracefulExitError()
        with patch.object(agent, "_shutdown"):
            agent.run()


if __name__ == "__main__":
    raise RuntimeError(
        "This test is not currently used and should be "
        "enabled in discover_tests.py if required."
    )



# Owner(s): ["module: onnx"]
# flake8: noqa: B950
"""Test op correctness by comparing with PyTorch results.

## Usage

1. Set the env var CATCH_ORT_SEGFAULT to catch segfaults from ONNX Runtime.

## How to add a new operator test

This test use PyTorch's OpInfo mechanism to generate test cases for each operator.
You may find all OpInfos in https://github.com/pytorch/pytorch/blob/7ec0d6f006fdd2c9b978dc6aa4923144684a3f51/torch/testing/_internal/common_methods_invocations.py#L8804

1. To enable test cases for an operator
    Add a `TorchLibOpInfo` entry to `TORCH_LIB_OPINFO` in `ops_test_data.py`.
    Specify `complex` if the function is designed for complex inputs.

    The `op_info_name` in `TorchLibOpInfo` needs to be unique in the TORCH_LIB_OPINFO
    list, but complex=True ops can share the same name with non-complex ops
    because they are tested separately.

2. Add `.skip` and/or `.xfail` to skip or xfail tests.
    Prefer xfail over skip when possible because that allows us to monitor the behavior
    and update the test will it passes.

    2a. If a test is now failing because of xpass, because some previous errors
    are now fixed, removed the corresponding xfail.

3. If sample inputs of the OpInfo needs to be adjusted to fit the aten signature, create an input
wrangler function. See `_mean_input_wrangler` for an example.

4. To test different ONNX functions that are registered as overloads of the same
    op, use `ops_test_common.duplicate_opinfo` to create new OpInfo with new names and map each
    to one overload.
"""

from __future__ import annotations

import copy
import dataclasses
import functools
from typing import Any, TYPE_CHECKING
from typing_extensions import Self

import numpy as np
import ops_test_common

import torch
from torch.onnx._internal.exporter._torchlib.ops import core as core_ops, nn as nn_ops
from torch.testing._internal import common_methods_invocations
from torch.testing._internal.opinfo import definitions as opinfo_definitions


if TYPE_CHECKING:
    from collections.abc import Callable, Collection


# Create a copy of the op_db to modify
OPS_DB = copy.deepcopy(common_methods_invocations.op_db)

# Append extra op_db into the op database for testing
OPS_DB.extend(opinfo_definitions.signal.op_db)


@dataclasses.dataclass
class TorchLibOpInfo:
    """A dataclass to store the information to test an torchlib op."""

    # The name of the op_info, e.g. "add"
    op_info_name: str
    # The torchlib ONNX Function to test
    op: Callable[..., Any]
    # The input wrangler function to adjust the input to fit the aten signature
    input_wrangler: (
        Callable[[list[Any], dict[str, Any]], tuple[list[Any], dict[str, Any]]] | None
    ) = None
    # Whether the op is non-deterministic
    nondeterministic: bool = False
    # Whether to compare the shape only for the output[index]
    # For example: (1,2) means compare value for output[0] and shape for output[1] and [2]
    # We may be able to combine this with the nondeterministic option
    compare_shape_only_for_output: tuple[int, ...] = ()
    # Whether the function is designed for complex inputs
    complex: bool = False
    # The ONNX opset version in which the function was introduced.
    # Its specifies the minimum ONNX opset version required to use the function.
    # It ensures that the function is only used when the target ONNX opset version
    # is compatible. For example, if `opset_introduced=20`, the function will only
    # be used when exporting to ONNX models targeting opset version 20 or higher.
    opset_introduced: int = 18
    # The acceptable tolerance of the inference result difference between PyTorch and ORT.
    # Format: {dtype: (rtol, atol)}.
    # For example: {torch.float16: (1e-3, 1e-3)}
    tolerance: dict[torch.dtype, tuple[float, float]] = dataclasses.field(
        default_factory=dict
    )
    # Expected skips or fails for the test and/or subtests
    skips_or_fails: list[ops_test_common.DecorateMeta] = dataclasses.field(
        default_factory=list
    )

    def get_tolerance(self, dtype: torch.dtype) -> tuple[float | None, float | None]:
        """Returns the (rtol, atol) tolerance for the given dtype."""
        if (tolerance := self.tolerance.get(dtype)) is not None:
            return tolerance

        # Use the PyTorch default if not specified
        # https://pytorch.org/docs/stable/testing.html
        return (None, None)

    def skip(
        self,
        variant_name: str = "",
        *,
        reason: str,
        dtypes: Collection[torch.dtype] | None = None,
        device_type: str | None = None,
        matcher: Callable[[Any], Any] | None = None,
        enabled_if: bool = True,
        test_class_name: str | None = None,
    ) -> Self:
        """Skips an OpInfo test.

        Args:
            variant_name: Optional OpInfo variant_test_name.
            reason: The reason for skipping.
            dtypes: The dtypes to skip.
            device_type: Device type. E.g. "cpu", "cuda".
            matcher: A function that matches the test sample input. It is used only when
                the skip is in the SKIP_XFAIL_SUBTESTS list.
            enabled_if: Whether the skip is enabled.
            test_class_name: The test class name to apply the skip to. If None, the skip
                is applied to all test classes.
        """
        self.skips_or_fails.append(
            ops_test_common.skip(
                self.op_info_name,
                variant_name,
                reason=reason,
                dtypes=dtypes,
                device_type=device_type,
                matcher=matcher,
                enabled_if=enabled_if,
                test_class_name=test_class_name,
            )
        )
        return self

    def xfail(
        self,
        variant_name: str = "",
        *,
        reason: str,
        dtypes: Collection[torch.dtype] | None = None,
        device_type: str | None = None,
        matcher: Callable[[Any], Any] | None = None,
        enabled_if: bool = True,
        test_class_name: str | None = None,
    ) -> Self:
        """Expects an OpInfo test to fail.

        Args:
            variant_name: Optional OpInfo variant_test_name.
            reason: The reason for the failure.
            dtypes: The dtypes to expect the failure
            device_type: Device type. E.g. "cpu", "cuda"..
            matcher: A function that matches the test sample input. It is used only when
                the xfail is in the SKIP_XFAIL_SUBTESTS list.
            enabled_if: Whether the xfail is enabled.
            test_class_name: The test class name to apply the xfail to. If None, the
                xfail is applied to all test classes.
        """
        self.skips_or_fails.append(
            ops_test_common.xfail(
                self.op_info_name,
                variant_name,
                reason=reason,
                dtypes=dtypes,
                device_type=device_type,
                matcher=matcher,
                enabled_if=enabled_if,
                test_class_name=test_class_name,
            )
        )
        return self


# Modify this section ##########################################################


def _amin_amax_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    if "dim" not in kwargs:
        # Supply an empty dim to match the aten signature
        kwargs["dim"] = np.array([], dtype=np.int64)
    else:
        # Convert dim to a numpy array
        kwargs["dim"] = np.array(kwargs["dim"], dtype=np.int64).reshape((-1,))
    return args, kwargs


def _avg_pool_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    if "dim" not in kwargs:
        if len(args) > 6:
            kwargs["divisor_override"] = args.pop(6)
        if len(args) > 5:
            kwargs["count_include_pad"] = args.pop(5)
        if len(args) > 4:
            kwargs["ceil_mode"] = args.pop(4)
        if len(args) > 3:
            padding = args.pop(3)
            if isinstance(padding, np.ndarray):
                # Cannot using list(padding) here, because the element will be numpy.int64 instead of int
                padding = padding.tolist()
            kwargs["padding"] = padding
        if len(args) > 2:
            stride = args.pop(2)
            if isinstance(stride, np.ndarray):
                stride = stride.tolist()
            kwargs["stride"] = stride
        kernel_size = args.pop(1)
        if isinstance(kernel_size, np.ndarray):
            kernel_size = kernel_size.tolist()
        kwargs["kernel_size"] = kernel_size
    return args, kwargs


def _cross_entropy_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    if "reduction" in kwargs:
        reduction_vals = ["none", "mean", "sum"]
        value = kwargs["reduction"]
        idx = reduction_vals.index(value)
        kwargs["reduction"] = idx
    return args, kwargs


def _dropout_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    if "training" in kwargs:
        kwargs["train"] = kwargs["training"]
        kwargs.pop("training")
    return args, kwargs


def _einsum_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    # Swap the equation and tensors to revert the special handling in the OpInfo
    return [args[1], args[0]], kwargs


def _embedding_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    """Remove arguments not present in the aten op signature."""
    kwargs.pop("max_norm", None)
    kwargs.pop("norm_type", None)
    return args, kwargs


def _empty_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    """Remove arguments not present in the aten op signature."""
    kwargs.pop("requires_grad", None)
    return args, kwargs


def _grid_sample_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    # Convert string attribute to int as input
    inter_mode_options = {"bilinear": 0, "nearest": 1, "bicubic": 2}
    padding_mode_options = {"zeros": 0, "border": 1, "reflection": 2}
    args.append(inter_mode_options[kwargs["mode"]])
    args.append(padding_mode_options[kwargs["padding_mode"]])
    args.append(kwargs["align_corners"])
    kwargs.clear()
    return args, kwargs


def _im2col_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    # Move kernel_size, dilation, padding and stride from args to kwargs
    if len(args) == 5:
        # Handle stride
        stride = args.pop()
        if isinstance(stride, np.ndarray):  # convert stride to list[int]
            stride = stride.tolist()
        kwargs["stride"] = stride
        # Handle padding
        padding = args.pop()
        if isinstance(padding, np.ndarray):  # convert padding to list[int]
            padding = padding.tolist()
        kwargs["padding"] = padding
        # Handle dilation
        dilation = args.pop()
        if isinstance(dilation, np.ndarray):  # convert dilation to list[int]
            dilation = dilation.tolist()
        kwargs["dilation"] = dilation
    # Handle kernel_size
    kernel_size = args.pop()
    if isinstance(kernel_size, np.ndarray):  # convert kernel_size to list[int]
        kernel_size = kernel_size.tolist()
    kwargs["kernel_size"] = kernel_size

    return args, kwargs


def _index_put_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    args[1] = [np.array(elem) for elem in args[1]]
    return args, kwargs


def _max_pool_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    # Remove return_indices argument because this op doesn't accept it
    kwargs.pop("return_indices", None)
    return args, kwargs


def _mean_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    # Make the dims as tensor
    if "dim" in kwargs:
        kwargs["dim"] = np.array(kwargs["dim"], dtype=np.int64)
    return args, kwargs


def _mse_loss_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    if "reduction" in kwargs:
        reduction_vals = ["none", "mean", "sum"]  # [0,1,2], default=1
        value = kwargs["reduction"]
        idx = reduction_vals.index(value)
        kwargs["reduction"] = idx
    return args, kwargs


def _nll_loss_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    if "reduction" in kwargs:
        # aten_nll_loss can only accept integer argument instead of string
        reduction_vals = ["none", "mean", "sum"]
        value = kwargs["reduction"]
        kwargs["reduction"] = reduction_vals.index(value)
    return args, kwargs


def _nonzero_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    kwargs.pop("as_tuple", None)
    return args, kwargs


def _reflection_pad2d_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    args.pop(2)  # remove 'reflect' arg
    return args, kwargs


def _replication_pad2d_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    args.pop(2)  # remove 'replicate' arg
    return args, kwargs


def _replication_pad3d_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    args.pop(2)  # remove 'replicate' arg
    return args, kwargs


def _roll_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    if len(args) >= 3:
        if isinstance(args[2], np.ndarray):  # convert dims to list[int]
            # Change dims from args to kwargs to keep tuple/list type
            dims = args.pop(2)
            kwargs["dims"] = dims.tolist()
        elif isinstance(args[2], int):  # convert dims to list[int]
            dims = args.pop(2)
            kwargs["dims"] = []
            kwargs["dims"].append(dims)
    if len(args) >= 2:
        if isinstance(args[1], np.ndarray):  # convert shift to list[int]
            shifts = args.pop(1)
            kwargs["shifts"] = shifts.tolist()
        elif isinstance(args[1], int):
            shifts = args.pop(1)
            kwargs["shifts"] = []
            kwargs["shifts"].append(shifts)
    return args, kwargs


def _scalar_tensor_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    kwargs.pop("requires_grad", None)
    return args, kwargs


def _scatter_reduce_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    # Put the string into kwargs, otherwise FullGraph mode could not find get 'reduce' argument
    kwargs["reduce"] = args.pop(4)
    return args, kwargs


def _sum_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    if kwargs.get("dim") is not None:
        kwargs["dim"] = np.array(kwargs["dim"], dtype=np.int64)
    return args, kwargs


def _unflatten_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    args[1] = np.array(args[1], dtype=np.int64)
    return args, kwargs


def _where_input_wrangler(
    args: list[Any], kwargs: dict[str, Any]
) -> tuple[list[Any], dict[str, Any]]:
    # The aten::where op takes condition, x, y as inputs
    # Swap the first two inputs
    args[0], args[1] = args[1], args[0]
    return args, kwargs


# Ops to be tested for numerical consistency between onnx and pytorch
# Find the names of the OpInfos in torch/testing/_internal/common_methods_invocations.py
TESTED_TORCHLIB_OPS: tuple[TorchLibOpInfo, ...] = (
    TorchLibOpInfo("abs", core_ops.aten_abs),
    TorchLibOpInfo("abs", core_ops.aten_abs_complex, complex=True),
    TorchLibOpInfo("add", core_ops.aten_add, tolerance={torch.float16: (1e-3, 1e-3)}),
    TorchLibOpInfo("add", core_ops.aten_add_complex, complex=True),
    TorchLibOpInfo("gelu_op20", nn_ops.aten_gelu_opset20, opset_introduced=20),
    TorchLibOpInfo(
        "nn.functional.group_norm", nn_ops.aten_group_norm, opset_introduced=21
    ).skip(
        reason="ONNX Runtime does not support zero sized inputs for GroupNorm",
        matcher=lambda sample: sample.input.numel() == 0,
    ),
    TorchLibOpInfo(
        "nn.functional.rms_norm", nn_ops.aten_rms_norm, opset_introduced=23
    ).skip(
        reason="ONNX Runtime does not support <1d inputs or zero sized inputs for RMSNorm",
        matcher=lambda sample: len(sample.input.shape) < 2 or sample.input.numel() == 0,
    ),
)


ops_test_common.duplicate_opinfo(OPS_DB, "all", ("all_dim", "all_dims"))
ops_test_common.duplicate_opinfo(OPS_DB, "any", ("any_dim", "any_dims"))
ops_test_common.duplicate_opinfo(
    OPS_DB, "arange", ("arange_start", "arange_start_step")
)
ops_test_common.duplicate_opinfo(OPS_DB, "atleast_1d", ("atleast_1d_Sequence",))
ops_test_common.duplicate_opinfo(OPS_DB, "atleast_2d", ("atleast_2d_Sequence",))
ops_test_common.duplicate_opinfo(OPS_DB, "atleast_3d", ("atleast_3d_Sequence",))
ops_test_common.duplicate_opinfo(
    OPS_DB,
    "bitwise_left_shift",
    (
        "bitwise_left_shift_int8",
        "bitwise_left_shift_int16",
        "bitwise_left_shift_int32",
        "bitwise_left_shift_int64",
    ),
)
ops_test_common.duplicate_opinfo(
    OPS_DB,
    "bitwise_right_shift",
    (
        "bitwise_right_shift_int8",
        "bitwise_right_shift_int16",
        "bitwise_right_shift_int32",
        "bitwise_right_shift_int64",
    ),
)
ops_test_common.duplicate_opinfo(OPS_DB, "cat", ("concat", "concatenate"))
ops_test_common.duplicate_opinfo(OPS_DB, "clone", ("lift_fresh_copy",))
ops_test_common.duplicate_opinfo(OPS_DB, "diagonal", ("diagonal_bool",))
ops_test_common.duplicate_opinfo(OPS_DB, "div", ("div_mode", "div_mode_int"))
ops_test_common.duplicate_opinfo(OPS_DB, "ge", ("ge_bool",))
ops_test_common.duplicate_opinfo(OPS_DB, "gt", ("gt_bool",))
ops_test_common.duplicate_opinfo(OPS_DB, "index_put", ("index_put_bool",))
ops_test_common.duplicate_opinfo(OPS_DB, "le", ("le_bool",))
ops_test_common.duplicate_opinfo(OPS_DB, "lt", ("lt_bool",))
ops_test_common.duplicate_opinfo(OPS_DB, "max", ("max_dim",))
ops_test_common.duplicate_opinfo(OPS_DB, "maximum", ("maximum_bool",))
ops_test_common.duplicate_opinfo(OPS_DB, "mean", ("mean_dim",))
ops_test_common.duplicate_opinfo(OPS_DB, "min", ("min_dim",))
ops_test_common.duplicate_opinfo(OPS_DB, "minimum", ("minimum_bool",))
ops_test_common.duplicate_opinfo(
    OPS_DB,
    "nn.functional.pad",
    (
        "nn.functional.reflection_pad2d",
        "nn.functional.replication_pad2d",
        "nn.functional.replication_pad3d",
    ),
)
ops_test_common.duplicate_opinfo(OPS_DB, "nn.functional.gelu", ("gelu_op20",))
ops_test_common.duplicate_opinfo(
    OPS_DB,
    "nn.functional.scaled_dot_product_attention",
    ("nn.functional.scaled_dot_product_attention_bool_mask",),
)
ops_test_common.duplicate_opinfo(
    OPS_DB,
    "nn.functional.celu",
    ("nn.functional.celu_type_promoted",),
)
ops_test_common.duplicate_opinfo(
    OPS_DB, "ops.aten._log_softmax", ("ops.aten._log_softmax_half",)
)
ops_test_common.duplicate_opinfo(
    OPS_DB, "ops.aten._softmax", ("ops.aten._softmax_half",)
)
ops_test_common.duplicate_opinfo(OPS_DB, "prod", ("prod_dim_int",))
ops_test_common.duplicate_opinfo(OPS_DB, "round", ("round_decimals",))
ops_test_common.duplicate_opinfo(OPS_DB, "squeeze", ("squeeze_dim",))
ops_test_common.duplicate_opinfo(OPS_DB, "view_as_complex", ("view_as_complex_copy",))
ops_test_common.duplicate_opinfo(OPS_DB, "view_as_real", ("view_as_real_copy",))

# MARK: End edits here


# These ops are not deterministic, so we check shape and dtype only
NONDETERMINISTIC_OPS: frozenset[str] = frozenset(
    info.op_info_name for info in TESTED_TORCHLIB_OPS if info.nondeterministic
)

COMPARE_SHAPE_ONLY_OPS: dict[
    str,
    set,
] = {
    info.op_info_name: set(info.compare_shape_only_for_output)
    for info in TESTED_TORCHLIB_OPS
}

TORCHLIB_OPINFO_MAPPING: dict[
    str,
    TorchLibOpInfo,
] = {info.op_info_name: info for info in TESTED_TORCHLIB_OPS if not info.complex}

TESTED_OPS = frozenset(TORCHLIB_OPINFO_MAPPING)

EXPECTED_SKIPS_OR_FAILS: tuple[ops_test_common.DecorateMeta, ...] = tuple(
    functools.reduce(
        # Flatten the list
        lambda a, b: [*a, *b],
        [
            [meta for meta in info.skips_or_fails if meta.matcher is None]
            for info in TESTED_TORCHLIB_OPS
        ],
    )
)

SKIP_XFAIL_SUBTESTS: tuple[ops_test_common.DecorateMeta, ...] = tuple(
    functools.reduce(
        # Flatten the list
        lambda a, b: [*a, *b],
        [
            [meta for meta in info.skips_or_fails if meta.matcher is not None]
            for info in TESTED_TORCHLIB_OPS
        ],
    )
)

# MARK: Complex supported functions
COMPLEX_FUNCTION_MAPPING: dict[
    str,
    TorchLibOpInfo,
] = {info.op_info_name: info for info in TESTED_TORCHLIB_OPS if info.complex}


# Call dir(torch.ops.prims) and compare with entries in OPS_DB to create OpInfo for newly added prims ops
PRIMS_OPS_WITH_OP_INFO = (
    "abs",
    "acos",
    "acosh",
    "add",
    "amax",
    "amin",
    "as_strided",
    "as_strided_scatter",
    "asin",
    "asinh",
    "atan",
    "atan2",
    "atanh",
    "bitwise_and",
    "bitwise_not",
    "bitwise_or",
    "bitwise_xor",
    "cat",
    "ceil",
    "clone",
    "conj",
    "conj_physical",
    "cos",
    "cosh",
    "digamma",
    "div",
    "empty",
    "eq",
    "erf",
    "erfc",
    "exp",
    "exp2",
    "expm1",
    "fill",
    "floor",
    "fmax",
    "fmin",
    "fmod",
    "full",
    "full_like",
    "gcd",
    "ge",
    "gt",
    "hypot",
    "igamma",
    "igammac",
    "imag",
    "isfinite",
    "le",
    "lgamma",
    "log",
    "log10",
    "log1p",
    "log2",
    "lt",
    "maximum",
    "minimum",
    "mul",
    "ne",
    "neg",
    "nextafter",
    "normal",
    "pow",
    "prod",
    "real",
    "reciprocal",
    "remainder",
    "reshape",
    "round",
    "rsqrt",
    "scalar_tensor",
    "sign",
    "signbit",
    "sin",
    "sinh",
    "sqrt",
    "squeeze",
    "sub",
    "sum",
    "svd",
    "tan",
    "tanh",
    "transpose",
    "trunc",
    "uniform",
    "where",
)

for op in PRIMS_OPS_WITH_OP_INFO:
    # Duplicate opinfo for prim ops. The new names all start with "prims_". E.g. "abs" -> "prims_abs".
    ops_test_common.duplicate_opinfo_for_prims(OPS_DB, op)

# Duplicate cases where the prims op name is different from the torch op name
ops_test_common.duplicate_opinfo_for_prims(OPS_DB, "i0", "bessel_i0")
ops_test_common.duplicate_opinfo_for_prims(OPS_DB, "special.bessel_j0", "bessel_j0")
ops_test_common.duplicate_opinfo_for_prims(OPS_DB, "special.bessel_j1", "bessel_j1")
ops_test_common.duplicate_opinfo_for_prims(OPS_DB, "special.erfcx", "erfcx")
ops_test_common.duplicate_opinfo_for_prims(OPS_DB, "special.i0e", "bessel_i0e")
ops_test_common.duplicate_opinfo_for_prims(OPS_DB, "special.i1", "bessel_i1")
ops_test_common.duplicate_opinfo_for_prims(OPS_DB, "special.i1e", "bessel_i1e")
ops_test_common.duplicate_opinfo_for_prims(OPS_DB, "special.ndtri", "ndtri")
ops_test_common.duplicate_opinfo_for_prims(
    OPS_DB, "special.spherical_bessel_j0", "spherical_bessel_j0"
)
ops_test_common.duplicate_opinfo_for_prims(OPS_DB, "special.zeta", "zeta")

OP_WITH_SKIPPED_XFAIL_SUBTESTS = frozenset(meta.op_name for meta in SKIP_XFAIL_SUBTESTS)
ALL_OPS_IN_DB = frozenset(op_info.name for op_info in OPS_DB)
# Assert all ops in OPINFO_FUNCTION_MAPPING are in the OPS_DB
if not TESTED_OPS.issubset(ALL_OPS_IN_DB):
    raise AssertionError(f"{TESTED_OPS - ALL_OPS_IN_DB} not in OPS_DB")
if not NONDETERMINISTIC_OPS.issubset(TESTED_OPS):
    raise AssertionError(f"{NONDETERMINISTIC_OPS - TESTED_OPS} not in TESTED_OPS")



# Owner(s): ["module: dynamo"]

from torch._dynamo import config
from torch._dynamo.testing import make_test_cls_with_patches


try:
    from . import test_export
except ImportError:
    import test_export


test_classes = {}


def make_dynamic_cls(cls):
    suffix = "_inline_and_install"

    cls_prefix = "InlineAndInstall"

    test_class = make_test_cls_with_patches(
        cls,
        cls_prefix,
        suffix,
        (config, "install_free_tensors", True),
        xfail_prop="_expected_failure_inline_and_install",
    )

    test_classes[test_class.__name__] = test_class
    # REMOVING THIS LINE WILL STOP TESTS FROM RUNNING
    globals()[test_class.__name__] = test_class
    test_class.__module__ = __name__
    return test_class


tests = [
    test_export.ExportTests,
]
for test in tests:
    make_dynamic_cls(test)
del test

if __name__ == "__main__":
    from torch._dynamo.test_case import run_tests

    run_tests()



"""Backward compatibility module for torch.onnx.symbolic_opset20."""

from __future__ import annotations


__all__: list[str] = []

from torch.onnx._internal.torchscript_exporter.symbolic_opset20 import *  # noqa: F401,F403
